from __future__ import annotations
from pathlib import Path
from typing import Any, DefaultDict, Dict, Optional, Union
import logging
from openghg.store.base import BaseStore
from xarray import Dataset

logger = logging.getLogger("openghg.store")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler

__all__ = ["EulerianModel"]


# TODO: Currently built around these keys but will probably need more unique distiguishers for different setups
# model name
# species
# date (start_date)
# ...
# setup (included as option for now)


class EulerianModel(BaseStore):
    """This class is used to process Eulerian model data"""

    _data_type = "eulerian_model"
    _root = "EulerianModel"
    _uuid = "63ff2365-3ba2-452a-a53d-110140805d06"
    _metakey = f"{_root}/uuid/{_uuid}/metastore"

    def read_file(
        self,
        filepath: Union[str, Path],
        model: str,
        species: str,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        setup: Optional[str] = None,
        if_exists: str = "auto",
        save_current: str = "auto",
        overwrite: bool = False,
        force: bool = False,
        compressor: Optional[Any] = None,
        filters: Optional[Any] = None,
        chunks: Optional[Dict] = None,
        optional_metadata: Optional[Dict] = None,
    ) -> Dict:
        """Read Eulerian model output

        Args:
            filepath: Path of Eulerian model species output
            model: Eulerian model name
            species: Species name
            start_date: Start date (inclusive) associated with model run
            end_date: End date (exclusive) associated with model run
            setup: Additional setup details for run
            if_exists: What to do if existing data is present.
                - "auto" - checks new and current data for timeseries overlap
                   - adds data if no overlap
                   - raises DataOverlapError if there is an overlap
                - "new" - just include new data and ignore previous
                - "combine" - replace and insert new data into current timeseries
            save_current: Whether to save data in current form and create a new version.
                - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
                - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
                - "n" / "no" - Allow current data to updated / deleted
            overwrite: Deprecated. This will use options for if_exists="new".
            force: Force adding of data even if this is identical to data stored.
            compressor: A custom compressor to use. If None, this will default to
                `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
                See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
            filters: Filters to apply to the data on storage, this defaults to no filtering. See
                https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
            chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
                for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
                See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking.
                To disable chunking pass in an empty dictionary.
            optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
        """
        # TODO: As written, this currently includes some light assumptions that we're dealing with GEOSChem SpeciesConc format.
        # May need to split out into multiple modules (like with ObsSurface) or into separate retrieve functions as needed.

        from collections import defaultdict
        from openghg.util import (
            clean_string,
            timestamp_now,
            timestamp_tzaware,
            check_if_need_new_version,
        )
        from pandas import Timestamp as pd_Timestamp
        from xarray import open_dataset

        model = clean_string(model)
        species = clean_string(species)
        start_date = clean_string(start_date)
        end_date = clean_string(end_date)
        setup = clean_string(setup)

        if overwrite and if_exists == "auto":
            logger.warning(
                "Overwrite flag is deprecated in preference to `if_exists` (and `save_current`) inputs."
                "See documentation for details of these inputs and options."
            )
            if_exists = "new"

        # Making sure new version will be created by default if force keyword is included.
        if force and if_exists == "auto":
            if_exists = "new"

        new_version = check_if_need_new_version(if_exists, save_current)

        filepath = Path(filepath)

        _, unseen_hashes = self.check_hashes(filepaths=filepath, force=force)

        if not unseen_hashes:
            return {}

        filepath = next(iter(unseen_hashes.values()))

        if chunks is None:
            chunks = {}

        with open_dataset(filepath).chunk(chunks) as em_data:
            # Check necessary 4D coordinates are present and rename if necessary (for consistency)
            check_coords = {
                "time": ["time"],
                "lat": ["lat", "latitude"],
                "lon": ["lon", "longitude"],
                "lev": ["lev", "level", "layer", "sigma_level"],
            }
            for name, coord_options in check_coords.items():
                for coord in coord_options:
                    if coord in em_data.coords:
                        break
                else:
                    raise ValueError(f"Input data must contain one of '{coord_options}' co-ordinate")
                if name != coord:
                    logger.info(f"Renaming co-ordinate '{coord}' to '{name}'")
                    em_data = em_data.rename({coord: name})

            attrs = em_data.attrs

            # author_name = "OpenGHG Cloud"
            # em_data.attrs["author"] = author_name

            metadata = {}
            metadata.update(attrs)

            metadata["model"] = model
            metadata["species"] = species
            metadata["processed"] = str(timestamp_now())
            metadata["data_type"] = "eulerian_model"

            if start_date is None:
                if len(em_data["time"]) > 1:
                    start_date = str(timestamp_tzaware(em_data.time[0].values))
                else:
                    try:
                        start_date = attrs["simulation_start_date_and_time"]
                    except KeyError:
                        raise Exception("Unable to derive start_date from data, please provide as an input.")
                    else:
                        start_date = timestamp_tzaware(start_date)
                        start_date = str(start_date)

            if end_date is None:
                if len(em_data["time"]) > 1:
                    end_date = str(timestamp_tzaware(em_data.time[-1].values))
                else:
                    try:
                        end_date = attrs["simulation_end_date_and_time"]
                    except KeyError:
                        raise Exception("Unable to derive `end_date` from data, please provide as an input.")
                    else:
                        end_date = timestamp_tzaware(end_date)
                        end_date = str(end_date)

            date = str(pd_Timestamp(start_date).date())

            metadata["date"] = date
            metadata["start_date"] = start_date
            metadata["end_date"] = end_date

            metadata["max_longitude"] = round(float(em_data["lon"].max()), 5)
            metadata["min_longitude"] = round(float(em_data["lon"].min()), 5)
            metadata["max_latitude"] = round(float(em_data["lat"].max()), 5)
            metadata["min_latitude"] = round(float(em_data["lat"].min()), 5)

            history = metadata.get("history")
            if history is None:
                history = ""
            metadata["history"] = history + f" {str(timestamp_now())} Processed onto OpenGHG cloud"

            key = "_".join((model, species, date))

            model_data: DefaultDict[str, Dict[str, Union[Dict, Dataset]]] = defaultdict(dict)
            model_data[key]["data"] = em_data
            model_data[key]["metadata"] = metadata

            required = ("model", "species", "date")

            if optional_metadata:
                common_keys = set(required) & set(optional_metadata.keys())

                if common_keys:
                    raise ValueError(
                        f"The following optional metadata keys are already present in required keys: {', '.join(common_keys)}"
                    )
                else:
                    for key, parsed_data in model_data.items():
                        parsed_data["metadata"].update(optional_metadata)

            data_type = "eulerian_model"
            datasource_uuids = self.assign_data(
                data=model_data,
                if_exists=if_exists,
                new_version=new_version,
                data_type=data_type,
                required_keys=required,
                compressor=compressor,
                filters=filters,
            )

            # TODO: MAY NEED TO ADD BACK IN OR CAN DELETE
            # update_keys = ["start_date", "end_date", "latest_version"]
            # model_data = update_metadata(
            #     data_dict=model_data, uuid_dict=datasource_uuids, update_keys=update_keys
            # )

            # em_store.add_datasources(
            #     uuids=datasource_uuids, data=model_data, metastore=metastore, update_keys=update_keys
            # )

            # Record the file hash in case we see this file again
            self.store_hashes(unseen_hashes)

            return datasource_uuids



# This holds store metadata for now
from typing import Dict
from openghg.store.base import BaseStore


def data_class_info() -> Dict:
    """Return storage class information for each data type.

    Returns:
        dict: Dictionary of storage class names and UUIDs
    """
    # TODO - move this to read from object store config file?

    return {
        "surface": {"_root": "ObsSurface", "_uuid": "da0b8b44-6f85-4d3c-b6a3-3dde34f6dea1"},
        "column": {"_root": "ObsColumn", "_uuid": "5c567168-0287-11ed-9d0f-e77f5194a415"},
        "flux": {"_root": "Flux", "_uuid": "c5c88168-0498-40ac-9ad3-949e91a30872"},
        "footprints": {"_root": "Footprints", "_uuid": "62db5bdf-c88d-4e56-97f4-40336d37f18c"},
        "boundary_conditions": {
            "_root": "BoundaryConditions",
            "_uuid": "4e787366-be91-4fc5-ad1b-4adcb213d478",
        },
        "eulerian_model": {"_root": "EulerianModel", "_uuid": "63ff2365-3ba2-452a-a53d-110140805d06"},
    }


def get_data_class(data_type: str) -> type[BaseStore]:
    """Return data class corresponding to given data type.

    Args:
        data_type: one of "surface", "column", "flux", "footprints",
    "boundary_conditions", or "eulerian_model"

    Returns:
        Data class, one of `ObsSurface`, `ObsColumn`, `Flux`, `EulerianModel`,
    `Footprints`, `BoundaryConditions`.
    """
    try:
        data_class = BaseStore._registry[data_type]
    except KeyError:
        raise ValueError(f"No data class for data type {data_type}.")
    else:
        return data_class



""" Helper functions to provide datapaths etc used in the tutorial notebooks

"""

import contextlib
import os
import shutil
import tarfile
import warnings
from pathlib import Path
from typing import List, Union
import logging
from openghg.standardise import standardise_footprint, standardise_flux, standardise_bc

logger = logging.getLogger("openghg.tutorial")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


def populate_footprint_data() -> None:
    """Adds all footprint data to the tutorial object store

    Returns:
        None
    """
    populate_footprint_inert()
    populate_footprint_co2()


def populate_footprint_inert() -> None:
    """Populates the tutorial object store with inert footprint data

    Returns:
        None
    """
    use_tutorial_store()

    tac_fp_inert = (
        "https://github.com/openghg/example_data/raw/main/footprint/tac_footprint_inert_201607.tar.gz"
    )

    tac_inert_path = retrieve_example_data(url=tac_fp_inert)[0]

    logger.info("Standardising footprint data...")
    # TODO - GJ - 2022-10-05 - This feels messy, how can we do this in a neater way?
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        with open(os.devnull, "w") as devnull:
            with contextlib.redirect_stdout(devnull):
                site = "TAC"
                # height = "100m"
                inlet = "100m"
                domain = "EUROPE"
                model = "NAME"

                standardise_footprint(
                    filepath=tac_inert_path, site=site, inlet=inlet, domain=domain, model=model
                )


def populate_footprint_co2() -> None:
    """Populates the tutorial object store with footprints data from the
    example data repository.

    Returns:
        None
    """
    tac_fp_co2 = "https://github.com/openghg/example_data/raw/main/footprint/tac_footprint_co2_201707.tar.gz"

    logger.info("Retrieving example data...")
    tac_co2_path = retrieve_example_data(url=tac_fp_co2)[0]

    logger.info("Standardising footprint data...")
    # TODO - GJ - 2022-10-05 - This feels messy, how can we do this in a neater way?
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        with open(os.devnull, "w") as devnull:
            with contextlib.redirect_stdout(devnull):
                site = "TAC"
                domain = "EUROPE"
                species = "co2"
                # height = "185m"
                inlet = "185m"
                model = "NAME"
                met_model = "UKV"

                standardise_footprint(
                    filepath=tac_co2_path,
                    site=site,
                    inlet=inlet,
                    domain=domain,
                    model=model,
                    met_model=met_model,
                    species=species,
                )

    logger.info("Done.")


def populate_flux_data() -> None:
    """Populate the tutorial store with flux data

    Returns:
        None
    """
    populate_flux_ch4()
    populate_flux_co2()


def populate_flux_co2() -> None:
    """Populate the tutorial object store with CO2 flux data

    Returns:
        None
    """
    co2_flux_eur = "https://github.com/openghg/example_data/raw/main/flux/co2-flux_EUROPE_2017.tar.gz"
    co2_flux_paths = retrieve_example_data(url=co2_flux_eur)

    source_natural = "natural"
    source_fossil = "ff-edgar-bp"

    flux_file_natural = [filename for filename in co2_flux_paths if source_natural in str(filename)][0]
    flux_file_ff = [filename for filename in co2_flux_paths if source_fossil in str(filename)][0]

    domain = "EUROPE"
    species = "co2"

    source_natural = "natural"
    source_fossil = "ff-edgar-bp"

    standardise_flux(
        filepath=flux_file_natural,
        species=species,
        source=source_natural,
        domain=domain,
        time_resolved=True,
    )
    standardise_flux(filepath=flux_file_ff, species=species, source=source_fossil, domain=domain)


def populate_flux_ch4() -> None:
    """Populates the tutorial object store with flux data from the
    example data repository.

    Returns:
        None
    """
    use_tutorial_store()

    logger.info("Retrieving data...")
    eur_2016_flux = "https://github.com/openghg/example_data/raw/main/flux/ch4-ukghg-all_EUROPE_2016.tar.gz"
    flux_data = retrieve_example_data(url=eur_2016_flux)

    source_waste = "waste"
    source_energyprod = "energyprod"

    flux_data_waste = [filename for filename in flux_data if source_waste in str(filename)][0]
    flux_data_energyprod = [filename for filename in flux_data if source_energyprod in str(filename)][0]

    logger.info("Standardising flux...")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        with open(os.devnull, "w") as devnull:
            with contextlib.redirect_stdout(devnull):
                domain = "EUROPE"
                species = "ch4"

                standardise_flux(
                    filepath=flux_data_waste, species=species, source=source_waste, domain=domain
                )
                standardise_flux(
                    filepath=flux_data_energyprod,
                    species=species,
                    source=source_energyprod,
                    domain=domain,
                )

    logger.info("Done.")


def populate_bc() -> None:
    """ """
    populate_bc_ch4()


def populate_bc_ch4() -> None:
    """Populates the tutorial object store with boundary conditions data from the
    example data repository.

    Returns:
        None
    """
    use_tutorial_store()

    logger.info("Retrieving data...")
    eur_2016_bc = (
        "https://github.com/openghg/example_data/raw/main/boundary_conditions/ch4_EUROPE_201607.tar.gz"
    )
    bc_data_path = retrieve_example_data(url=eur_2016_bc)[0]

    bc_input = "CAMS"
    domain = "EUROPE"
    species = "ch4"

    logger.info("Standardising boundary conditions...")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        with open(os.devnull, "w") as devnull:
            with contextlib.redirect_stdout(devnull):
                standardise_bc(filepath=bc_data_path, bc_input=bc_input, species=species, domain=domain)

    logger.info("Done.")


def populate_surface_data() -> None:
    """Populates the tutorial object store with surface measurement data from the
    example data repository.

    Returns:
        None
    """
    from openghg.standardise import standardise_surface

    use_tutorial_store()

    bsd_data = "https://github.com/openghg/example_data/raw/main/timeseries/bsd_example.tar.gz"
    tac_data = "https://github.com/openghg/example_data/raw/main/timeseries/tac_example.tar.gz"
    capegrim_data = "https://github.com/openghg/example_data/raw/main/timeseries/capegrim_example.tar.gz"

    logger.info("Retrieving example data...")
    bsd_paths = retrieve_example_data(url=bsd_data)
    tac_paths = retrieve_example_data(url=tac_data)
    capegrim_paths = sorted(retrieve_example_data(url=capegrim_data))

    # Create the tuple required
    capegrim_tuple = (capegrim_paths[0], capegrim_paths[1])

    logger.info("Standardising data...")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        with open(os.devnull, "w") as devnull:
            with contextlib.redirect_stdout(devnull):
                standardise_surface(filepaths=bsd_paths, source_format="crds", site="bsd", network="decc")
                standardise_surface(filepaths=tac_paths, source_format="crds", site="tac", network="decc")
                standardise_surface(
                    filepaths=capegrim_tuple,
                    instrument="medusa",
                    source_format="gcwerks",
                    site="cgo",
                    network="agage",
                )

    logger.info("Done.")


def download_edgar_data() -> Path:
    """
    Download edgar data to tutorial store to be used within parse_edgar transform
    tutorial.

    This is currently a limited subset of v6.0 CH4 data (2014-2015)

    TODO: Upgrade to use v7.0 data when this has been checked and added into workflow.
    """

    use_tutorial_store()

    edgar_v60_database = "https://github.com/openghg/example_data/raw/main/databases/TOTALS_nc.tar.gz"

    logger.info("Retrieving example database...")
    edgar_database_path = retrieve_example_data(url=edgar_v60_database)[0]
    logger.info("Done.")

    return edgar_database_path


def use_tutorial_store() -> None:
    """Sets an environment variable telling OpenGHG to use a
    temporary object store. This sets the store to be
    the result of tempfile.gettempdir() / openghg_temp_store.
    To tidy up this store use the clean_tutorial_store function.

    Returns:
        None
    """
    os.environ["OPENGHG_TUT_STORE"] = "1"


def example_extract_path() -> Path:
    """Return the path to folder containing the extracted example files

    Returns:
        None
    """
    from openghg.objectstore import get_tutorial_store_path

    return Path(get_tutorial_store_path(), "extracted_files")


def clear_example_cache() -> None:
    """Removes the file cache created when running the tutorials.

    Returns:
        None
    """
    from openghg.objectstore import get_tutorial_store_path

    example_cache_path = get_tutorial_store_path() / "example_cache"
    extracted_examples = example_extract_path()

    if example_cache_path.exists():
        shutil.rmtree(example_cache_path, ignore_errors=True)
        shutil.rmtree(extracted_examples, ignore_errors=True)


def retrieve_example_obspack(extract_dir: Union[str, Path, None] = None) -> Path:
    """Retrieves our example ObsPack dataset, extracts it and returns the path to the folder.

    Args:
        url: URL to retrieve.
        extract_dir: Folder to extract example tarballs to
    Returns:
        Path: Path to directory
    """
    url = "https://github.com/openghg/example_data/raw/main/obspack/obspack_ch4_example.tar.gz"
    files = retrieve_example_data(url=url, extract_dir=extract_dir)
    return files[0].parent


def retrieve_example_data(url: str, extract_dir: Union[str, Path, None] = None) -> List[Path]:
    """Retrieve data from the OpenGHG example data repository, cache the downloaded data,
    extract the data and return the filepaths of the extracted files.

    Args:
        url: URL to retrieve.
        extract_dir: Folder to extract example tarballs to
    Returns:
        list: List of filepaths
    """
    from openghg.objectstore import get_tutorial_store_path
    from openghg.util import download_data, parse_url_filename

    use_tutorial_store()

    # Check we're getting a tar
    output_filename = parse_url_filename(url=url)

    suffixes = Path(output_filename).suffixes
    if ".tar" not in suffixes:
        raise ValueError("This function can only currently works with tar files.")

    example_cache_path = get_tutorial_store_path() / "example_cache"

    if not example_cache_path.exists():
        example_cache_path.mkdir(parents=True)

    # cache_record = example_cache_path / "cache_record.json"
    download_path = Path(example_cache_path).joinpath(output_filename)

    # cache_exists = cache_record.is_file()

    # if cache_exists:
    #     cache_data = json.loads(cache_record.read_text())

    #     try:
    #         cached_datapath = Path(cache_data[output_filename])
    #     except KeyError:
    #         cache_data[output_filename] = str(download_path)
    #     else:
    #         return unpack_example_archive(archive_path=cached_datapath, extract_dir=extract_dir)
    # else:
    #     cache_data = {}
    #     cache_data[output_filename] = str(download_path)

    download_data(url=url, filepath=download_path)

    if not download_path.exists():
        raise ValueError("Unable to download file. Please check the URL is correct.")

    # # Make sure we still have all the files in the cache we expect to
    # checked_cache = {}
    # for filename, path in cache_data.items():
    #     if Path(path).exists():
    #         checked_cache[filename] = path

    # cache_record.write_text(json.dumps(checked_cache))

    return unpack_example_archive(archive_path=download_path, extract_dir=extract_dir)


def unpack_example_archive(archive_path: Path, extract_dir: Union[str, Path, None] = None) -> List[Path]:
    """Unpacks an tar file to a temporary folder, or extract_dir if given.
    Returns the filepath(s) of the objects.

    Returns:
        list: List of filepaths
    """
    from openghg.tutorial import example_extract_path

    if extract_dir is None:
        extract_dir = example_extract_path()

    with tarfile.open(archive_path) as tar:
        filenames = [f.name for f in tar.getmembers()]

    shutil.unpack_archive(filename=archive_path, extract_dir=extract_dir)

    extracted_filepaths = [Path(extract_dir, str(fname)) for fname in filenames]

    return extracted_filepaths


def clear_tutorial_store() -> None:
    """Delete the contents of the tutorial object store

    Returns:
        None
    """
    from openghg.objectstore import get_tutorial_store_path

    path = get_tutorial_store_path()

    shutil.rmtree(path=path, ignore_errors=True)

    logger.info(f"Tutorial store at {path} cleared.")



from __future__ import annotations
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union, cast
import warnings
import numpy as np
from openghg.store import DataSchema
from openghg.store.base import BaseStore
from openghg.store.storage import ChunkingSchema
from openghg.util import species_lifetime, synonyms
from xarray import Dataset

__all__ = ["Footprints"]

logger = logging.getLogger("openghg.store")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


class Footprints(BaseStore):
    """This class is used to process footprints model output"""

    _data_type = "footprints"
    _root = "Footprints"
    _uuid = "62db5bdf-c88d-4e56-97f4-40336d37f18c"
    _metakey = f"{_root}/uuid/{_uuid}/metastore"

    def read_data(self, binary_data: bytes, metadata: Dict, file_metadata: Dict) -> Optional[Dict]:
        """Ready a footprint from binary data

        Args:
            binary_data: Footprint data
            metadata: Dictionary of metadata
            file_metadat: File metadata
        Returns:
            dict: UUIDs of Datasources data has been assigned to
        """
        raise NotImplementedError("This branch doesn't currently support cloud.")
        # with TemporaryDirectory() as tmpdir:
        #     tmpdir_path = Path(tmpdir)

        #     try:
        #         filename = file_metadata["filename"]
        #     except KeyError:
        #         raise KeyError("We require a filename key for metadata read.")

        #     filepath = tmpdir_path.joinpath(filename)
        #     filepath.write_bytes(binary_data)

        #     return Footprints.read_file(filepath=filepath, **metadata)

    # @staticmethod
    # def read_data(binary_data: bytes, metadata: Dict, file_metadata: Dict) -> Dict:
    #     """Ready a footprint from binary data

    #     Args:
    #         binary_data: Footprint data
    #         metadata: Dictionary of metadata
    #         file_metadat: File metadata
    #     Returns:
    #         dict: UUIDs of Datasources data has been assigned to
    #     """
    #     from openghg.store import assign_data, infer_date_range, load_metastore, datasource_lookup

    #     fp = Footprints.load()

    #     # Load in the metadata store
    #     metastore = load_metastore(key=fp._metakey)

    #     sha1_hash = file_metadata["sha1_hash"]
    #     overwrite = metadata.get("overwrite", False)

    #     if sha1_hash in fp._file_hashes and not overwrite:
    #         print(
    #             f"This file has been uploaded previously with the filename : {fp._file_hashes[sha1_hash]} - skipping."
    #         )

    #     data_buf = BytesIO(binary_data)
    #     fp_data = open_dataset(data_buf)

    #     fp_time = fp_data.time
    #     period = metadata.get("period")
    #     continuous = metadata["continous"]
    #     high_spatial_res = metadata["high_spatial_res"]
    #     species = metadata["species"]
    #     filename = file_metadata["filename"]

    #     site = metadata["site"]
    #     domain = metadata["domain"]
    #     model = metadata["model"]
    #     height = metadata["height"]

    #     start_date, end_date, period_str = infer_date_range(
    #         fp_time, filepath=filename, period=period, continuous=continuous
    #     )

    #     metadata["start_date"] = str(start_date)
    #     metadata["end_date"] = str(end_date)
    #     metadata["time_period"] = period_str

    #     metadata["max_longitude"] = round(float(fp_data["lon"].max()), 5)
    #     metadata["min_longitude"] = round(float(fp_data["lon"].min()), 5)
    #     metadata["max_latitude"] = round(float(fp_data["lat"].max()), 5)
    #     metadata["min_latitude"] = round(float(fp_data["lat"].min()), 5)

    #     # TODO: Pull out links to underlying data format into a separate format function
    #     #  - high_spatial_res - data vars - "fp_low", "fp_high", coords - "lat_high", "lon_high"
    #     #  - high_time_res - data vars - "fp_HiTRes", coords - "H_back"

    #     metadata["spatial_resolution"] = "standard_spatial_resolution"

    #     if high_spatial_res:
    #         try:
    #             metadata["max_longitude_high"] = round(float(fp_data["lon_high"].max()), 5)
    #             metadata["min_longitude_high"] = round(float(fp_data["lon_high"].min()), 5)
    #             metadata["max_latitude_high"] = round(float(fp_data["lat_high"].max()), 5)
    #             metadata["min_latitude_high"] = round(float(fp_data["lat_high"].min()), 5)

    #             metadata["spatial_resolution"] = "high_spatial_resolution"
    #         except KeyError:
    #             raise KeyError("Expected high spatial resolution. Unable to find lat_high or lon_high data.")

    #     if species == "co2":
    #         # Expect co2 data to have high time resolution
    #         high_time_res = True

    #     metadata["time_resolution"] = "standard_time_resolution"

    #     if high_time_res:
    #         if "fp_HiTRes" in fp_data:
    #             metadata["time_resolution"] = "high_time_resolution"
    #         else:
    #             raise KeyError("Expected high time resolution. Unable to find fp_HiTRes data.")

    #     metadata["heights"] = [float(h) for h in fp_data.height.values]
    #     # Do we also need to save all the variables we have available in this footprints?
    #     metadata["variables"] = list(fp_data.keys())

    #     # if model_params is not None:
    #     #     metadata["model_parameters"] = model_params

    #     # Set the attributes of this Dataset
    #     fp_data.attrs = {"author": "OpenGHG Cloud", "processed": str(timestamp_now())}

    #     # This might seem longwinded now but will help when we want to read
    #     # more than one footprints at a time
    #     key = "_".join((site, domain, model, height))

    #     footprint_data: DefaultDict[str, Dict[str, Union[Dict, Dataset]]] = defaultdict(dict)
    #     footprint_data[key]["data"] = fp_data
    #     footprint_data[key]["metadata"] = metadata

    #     # These are the keys we will take from the metadata to search the
    #     # metadata store for a Datasource, they should provide as much detail as possible
    #     # to uniquely identify a Datasource
    #     required = ("site", "model", "height", "domain")
    #     lookup_results = datasource_lookup(metastore=metastore, data=footprint_data, required_keys=required)

    #     data_type = "footprints"
    #     datasource_uuids: Dict[str, Dict] = assign_data(
    #         data_dict=footprint_data,
    #         lookup_results=lookup_results,
    #         overwrite=overwrite,
    #         data_type=data_type,
    #     )

    #     fp.add_datasources(uuids=datasource_uuids, data=footprint_data, metastore=metastore)

    #     # Record the file hash in case we see this file again
    #     fp._file_hashes[sha1_hash] = filename

    #     fp.save()

    #     metastore.close()

    #     return datasource_uuids

    # def _store_data(data: Dataset, metadata: Dict):
    #     """ Takes an xarray Dataset

    #     Args:
    #         data: xarray Dataset
    #         metadata: Metadata dict
    #     Returns:

    #     """

    def read_file(
        self,
        filepath: Union[List, str, Path],
        site: str,
        domain: str,
        model: str,
        inlet: Optional[str] = None,
        height: Optional[str] = None,
        met_model: Optional[str] = None,
        species: Optional[str] = None,
        network: Optional[str] = None,
        period: Optional[Union[str, tuple]] = None,
        continuous: bool = True,
        chunks: Optional[Dict] = None,
        source_format: str = "acrg_org",
        retrieve_met: bool = False,
        high_spatial_resolution: bool = False,
        time_resolved: bool = False,
        high_time_resolution: bool = False,
        short_lifetime: bool = False,
        if_exists: str = "auto",
        save_current: str = "auto",
        overwrite: bool = False,
        force: bool = False,
        sort: bool = False,
        drop_duplicates: bool = False,
        compressor: Optional[Any] = None,
        filters: Optional[Any] = None,
        optional_metadata: Optional[Dict] = None,
    ) -> dict:
        """Reads footprints data files and returns the UUIDS of the Datasources
        the processed data has been assigned to

        Args:
            filepath: Path(s) of file(s) to standardise
            site: Site name
            domain: Domain of footprints
            model: Model used to create footprint (e.g. NAME or FLEXPART)
            inlet: Height above ground level in metres. Format 'NUMUNIT' e.g. "10m"
            height: Alias for inlet. One of height or inlet MUST be included.
            met_model: Underlying meteorlogical model used (e.g. UKV)
            species: Species name. Only needed if footprint is for a specific species e.g. co2 (and not inert)
            network: Network name
            period: Period of measurements. Only needed if this can not be inferred from the time coords
            chunks: Chunking schema to use when storing data.
            continuous: Whether time stamps have to be continuous.
            chunks: Chunk schema to use when storing data the NetCDF. It expects a dictionary of dimension name and chunk size,
                for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
            source_format : Type of data being input e.g. acrg_org
            retrieve_met: Whether to also download meterological data for this footprints area
            high_spatial_resolution : Indicate footprints include both a low and high spatial resolution.
            time_resolved: Indicate footprints are high time resolution (include H_back dimension)
                           Note this will be set to True automatically if species="co2" (Carbon Dioxide).
            high_time_resolution: This argument is deprecated and will be replaced in future versions with time_resolved.
            short_lifetime: Indicate footprint is for a short-lived species. Needs species input.
                            Note this will be set to True if species has an associated lifetime.
            if_exists: What to do if existing data is present.
                - "auto" - checks new and current data for timeseries overlap
                   - adds data if no overlap
                   - raises DataOverlapError if there is an overlap
                - "new" - just include new data and ignore previous
                - "combine" - replace and insert new data into current timeseries
            save_current: Whether to save data in current form and create a new version.
                - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
                - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
                - "n" / "no" - Allow current data to updated / deleted
            overwrite: Deprecated. This will use options for if_exists="new".
            force: Force adding of data even if this is identical to data stored.
            sort: Sort data in time dimension. We recommend NOT sorting footprint data unless necessary.
            drop_duplicates: Drop duplicate timestamps, keeping the first value
            compressor: A custom compressor to use. If None, this will default to
                `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
                See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
            filters: Filters to apply to the data on storage, this defaults to no filtering. See
                https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
            optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
        Returns:
            dict: UUIDs of Datasources data has been assigned to
        """
        from openghg.types import FootprintTypes

        from openghg.util import clean_string, format_inlet, check_if_need_new_version, load_footprint_parser

        if not isinstance(filepath, list):
            filepath = [filepath]

        if high_time_resolution:
            warnings.warn(
                "This argument is deprecated and will be replaced in future versions with time_resolved.",
                DeprecationWarning,
            )
            time_resolved = high_time_resolution

        # We wanted sorted Path objects
        filepath = sorted([Path(f) for f in filepath])

        site = clean_string(site)
        network = clean_string(network)
        domain = clean_string(domain)

        # Make sure `inlet` OR the alias `height` is included
        # Note: from this point only `inlet` variable should be used.
        if inlet is None and height is None:
            raise ValueError("One of inlet (or height) must be specified as an input")
        elif inlet is None:
            inlet = height

        # Try to ensure inlet is 'NUM''UNIT' e.g. "10m"
        inlet = clean_string(inlet)
        inlet = format_inlet(inlet)
        inlet = cast(str, inlet)

        # Ensure we have a value for species
        if species is None:
            species = "inert"
        else:
            species = clean_string(species)
            species = synonyms(species)

        # Ensure we have a clear missing value for met_model
        if met_model is None:
            met_model = "NOT_SET"
        else:
            met_model = clean_string(met_model)

        if network is not None:
            network = clean_string(network)

        try:
            source_format = FootprintTypes[source_format.upper()].value
        except KeyError:
            raise ValueError(f"Unknown data type {source_format} selected.")

        # Load the data retrieve object
        parser_fn = load_footprint_parser(source_format=source_format)

        # file_hash = hash_file(filepath=filepath)
        # if file_hash in self._file_hashes and not overwrite:
        if overwrite and if_exists == "auto":
            logger.warning(
                "Overwrite flag is deprecated in preference to `if_exists` (and `save_current`) inputs."
                "See documentation for details of these inputs and options."
            )
            if_exists = "new"

        # Making sure new version will be created by default if force keyword is included.
        if force and if_exists == "auto":
            if_exists = "new"

        new_version = check_if_need_new_version(if_exists, save_current)

        _, unseen_hashes = self.check_hashes(filepaths=filepath, force=force)

        if not unseen_hashes:
            return {}

        filepath = list(unseen_hashes.values())

        if not filepath:
            return {}

        # Do some housekeeping on the inputs
        if species == "co2":
            # Expect co2 data to have high time resolution
            if not time_resolved:
                logger.info("Updating time_resolved to True for co2 data")
                time_resolved = True

            if sort:
                logger.info(
                    "Sorting high time resolution data is very memory intensive, we recommend not sorting."
                )

        if short_lifetime:
            if species == "inert":
                raise ValueError(
                    "When indicating footprint is for short lived species, 'species' input must be included"
                )
        else:
            if species == "inert":
                lifetime = None
            else:
                lifetime = species_lifetime(species)
                if lifetime is not None:
                    # TODO: May want to add a check on length of lifetime here
                    logger.info("Updating short_lifetime to True since species has an associated lifetime")
                    short_lifetime = True

        chunks = self.check_chunks(
            filepaths=filepath,
            chunks=chunks,
            high_spatial_resolution=high_spatial_resolution,
            time_resolved=time_resolved,
            short_lifetime=short_lifetime,
        )

        # Define parameters to pass to the parser function
        # TODO: Update this to match against inputs for parser function.
        param = {
            "filepath": filepath,
            "site": site,
            "domain": domain,
            "model": model,
            "inlet": inlet,
            "met_model": met_model,
            "species": species,
            "network": network,
            "time_resolved": time_resolved,
            "high_spatial_resolution": high_spatial_resolution,
            "short_lifetime": short_lifetime,
            "period": period,
            "continuous": continuous,
            "chunks": chunks,
        }

        input_parameters: dict[Any, Any] = param.copy()

        # # TODO: Decide if we want to include details below / switch any parameters to be optional.
        # optional_keywords: dict[Any, Any] = {}

        # signature = inspect.signature(parser_fn)
        # fn_accepted_parameters = [param.name for param in signature.parameters.values()]

        # # Checks if optional parameters are present in function call and includes them else ignores its inclusion in input_parameters.
        # for param, param_value in optional_keywords.items():
        #     if param in fn_accepted_parameters:
        #         input_parameters[param] = param_value
        #     else:
        #         logger.warning(
        #             f"Input: '{param}' (value: {param_value}) is not being used as part of the standardisation process."
        #             f"This is not accepted by the current standardisation function: {parser_fn}"
        #         )

        footprint_data = parser_fn(**input_parameters)

        # Checking against expected format for footprints
        # Based on configuration (some user defined, some inferred)
        for split_data in footprint_data.values():
            fp_data = split_data["data"]
            Footprints.validate_data(
                fp_data,
                high_spatial_resolution=high_spatial_resolution,
                time_resolved=time_resolved,
                short_lifetime=short_lifetime,
            )

        if species == "co2" and sort is True:
            logger.info(
                "Sorting high time resolution data is very memory intensive, we recommend not sorting."
            )

        # These are the keys we will take from the metadata to search the
        # metadata store for a Datasource, they should provide as much detail as possible
        # to uniquely identify a Datasource
        required = (
            "site",
            "model",
            "inlet",
            "domain",
            "time_resolved",
            "high_spatial_resolution",
            "short_lifetime",
            "species",
            "met_model",
        )

        if optional_metadata:
            common_keys = set(required) & set(optional_metadata.keys())

            if common_keys:
                raise ValueError(
                    f"The following optional metadata keys are already present in required keys: {', '.join(common_keys)}"
                )
            else:
                for key, parsed_data in footprint_data.items():
                    parsed_data["metadata"].update(optional_metadata)

        data_type = "footprints"
        # TODO - filter options
        datasource_uuids = self.assign_data(
            data=footprint_data,
            if_exists=if_exists,
            new_version=new_version,
            data_type=data_type,
            required_keys=required,
            sort=sort,
            drop_duplicates=drop_duplicates,
            compressor=compressor,
            filters=filters,
        )

        # TODO: MAY NEED TO ADD BACK IN OR CAN DELETE
        # update_keys = ["start_date", "end_date", "latest_version"]
        # footprint_data = update_metadata(
        #     data_dict=footprint_data, uuid_dict=datasource_uuids, update_keys=update_keys
        # )

        # Record the file hash in case we see the file(s) again
        self.store_hashes(unseen_hashes)

        return datasource_uuids

    @staticmethod
    def schema(
        particle_locations: bool = True,
        high_spatial_resolution: bool = False,
        time_resolved: bool = False,
        high_time_resolution: bool = False,
        short_lifetime: bool = False,
    ) -> DataSchema:
        """
        Define schema for footprint Dataset.

        The returned schema depends on what the footprint represents,
        indicated using the keywords.
        By default, this will include "fp" variable but this will be superceded
        if high_spatial_resolution or time_resolved are specified.

        Args:
            particle_locations: Include 4-directional particle location variables:
                - "particle_location_[nesw]"
                and include associated additional dimensions ("height")
            high_spatial_resolution : Set footprint variables include high and low resolution options:
                - "fp_low"
                - "fp_high"
                and include associated additional dimensions ("lat_high", "lon_high").
            time_resolved: Set footprint variable to be high time resolution
                - "fp_HiTRes"
                and include associated dimensions ("H_back").
            high_time_resolution: This argument is deprecated and will be replaced in future versions with time_resolved.
            short_lifetime: Include additional particle age parameters for short lived species:
                - "mean_age_particles_[nesw]"

        Returns:
            DataSchema object describing this format.

        Note: In PARIS format the coordinate dimensions are ("latitude", "longitude") rather than ("lat", "lon")
            but given that all other openghg internal formats are ("lat", "lon"), we are currently keeping all
            footprint internal formats consistent with this.
        """

        # # Note: In PARIS format the coordinate dimensions are ("latitude", "longitude") rather than ("lat", "lon")
        # # but given that all other openghg internal formats are ("lat", "lon"), we are currently keeping the
        # # footprint internal format consistent with this.

        # Names of data variables and associated dimensions (as a tuple)
        data_vars: Dict[str, Tuple[str, ...]] = {}
        # Internal data types of data variables and coordinates
        dtypes = {
            "lat": np.floating,  # Covers np.float16, np.float32, np.float64 types
            "lon": np.floating,
            "time": np.datetime64,
        }

        if high_time_resolution:
            warnings.warn(
                "This argument is deprecated and will be replaced in future versions with time_resolved.",
                DeprecationWarning,
            )
            time_resolved = high_time_resolution

        if not time_resolved and not high_spatial_resolution:
            # Includes standard footprint variable
            data_vars["fp"] = ("time", "lat", "lon")
            dtypes["fp"] = np.floating

        if high_spatial_resolution:
            # Include options for high spatial resolution footprint
            # This includes footprint data on multiple resolutions

            data_vars["fp_low"] = ("time", "lat", "lon")
            data_vars["fp_high"] = ("time", "lat_high", "lon_high")

            dtypes["fp_low"] = np.floating
            dtypes["fp_high"] = np.floating

        if time_resolved:
            # Include options for high time resolution footprint (usually co2)
            # This includes a footprint data with an additional hourly back dimension
            data_vars["fp_HiTRes"] = ("time", "lat", "lon", "H_back")
            dtypes["fp_HiTRes"] = np.floating
            dtypes["H_back"] = np.number  # float or integer

        # Includes particle location directions - one for each regional boundary
        if particle_locations:
            data_vars["particle_locations_n"] = ("time", "lon", "height")
            data_vars["particle_locations_e"] = ("time", "lat", "height")
            data_vars["particle_locations_s"] = ("time", "lon", "height")
            data_vars["particle_locations_w"] = ("time", "lat", "height")

            dtypes["height"] = np.floating
            dtypes["particle_locations_n"] = np.floating
            dtypes["particle_locations_e"] = np.floating
            dtypes["particle_locations_s"] = np.floating
            dtypes["particle_locations_w"] = np.floating

        # TODO: Could also add check for meteorological + other data
        # "pressure", "wind_speed", "wind_direction", "PBLH"
        # "release_lon", "release_lat"

        # Include options for short lifetime footprints (short-lived species)
        # This includes additional particle ages (allow calculation of decay based on particle lifetimes)
        if short_lifetime:
            data_vars["mean_age_particles_n"] = ("time", "lon", "height")
            data_vars["mean_age_particles_e"] = ("time", "lat", "height")
            data_vars["mean_age_particles_s"] = ("time", "lon", "height")
            data_vars["mean_age_particles_w"] = ("time", "lat", "height")

            dtypes["mean_age_particles_n"] = np.floating
            dtypes["mean_age_particles_e"] = np.floating
            dtypes["mean_age_particles_s"] = np.floating
            dtypes["mean_age_particles_w"] = np.floating

        data_format = DataSchema(data_vars=data_vars, dtypes=dtypes)

        return data_format

    @staticmethod
    def validate_data(
        data: Dataset,
        particle_locations: bool = True,
        high_spatial_resolution: bool = False,
        time_resolved: bool = False,
        high_time_resolution: bool = False,
        short_lifetime: bool = False,
    ) -> None:
        """
        Validate data against Footprint schema - definition from
        Footprints.schema(...) method.

        Args:
            data : xarray Dataset in expected format

            See Footprints.schema() method for details on optional inputs.

        Returns:
            None

            Raises a ValueError with details if the input data does not adhere
            to the Footprints schema.
        """
        if high_time_resolution:
            warnings.warn(
                "This argument is deprecated and will be replaced in future versions with time_resolved.",
                DeprecationWarning,
            )
            time_resolved = high_time_resolution
        data_schema = Footprints.schema(
            particle_locations=particle_locations,
            high_spatial_resolution=high_spatial_resolution,
            time_resolved=time_resolved,
            short_lifetime=short_lifetime,
        )
        data_schema.validate_data(data)

    def chunking_schema(
        self,
        time_resolved: bool = False,
        high_time_resolution: bool = False,
        high_spatial_resolution: bool = False,
        short_lifetime: bool = False,
    ) -> ChunkingSchema:
        """
        Get chunking schema for footprint data.

        Args:
            time_resolved : Set footprint variable to be high time resolution.
            high_time_resolution: This argument is deprecated and will be replaced in future versions with time_resolved.
            high_spatial_resolution : Set footprint variables include high and low resolution options.
            short_lifetime: Include additional particle age parameters for short lived species.
        Returns:
            dict: Chunking schema for footprint data.
        """
        if high_spatial_resolution or short_lifetime:
            raise NotImplementedError(
                "Chunking schema for footprints with high spatial resolution or short lifetime is not currently set.\n"
                + "Using the default chunking schema."
            )

        # TODO - could these defaults be changed in the object store config maybe?

        if high_time_resolution:
            warnings.warn(
                "This argument is deprecated and will be replaced in future versions with time_resolved.",
                DeprecationWarning,
            )
            time_resolved = high_time_resolution
        if time_resolved:
            var = "fp_HiTRes"
            time_chunk_size = 24
            secondary_vars = ["lat", "lon", "H_back"]
        else:
            var = "fp"
            time_chunk_size = 480
            secondary_vars = ["lat", "lon"]

        return ChunkingSchema(variable=var, chunks={"time": time_chunk_size}, secondary_dims=secondary_vars)



from __future__ import annotations
import logging
from pathlib import Path
from typing import Any, Dict, Optional, Union

from numpy import ndarray

# from openghg.store import DataSchema
from openghg.store.base import BaseStore
from xarray import DataArray

ArrayType = Optional[Union[ndarray, DataArray]]

logger = logging.getLogger("openghg.store")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


class ObsColumn(BaseStore):
    """This class is used to process emissions / flux data"""

    _data_type = "column"
    _root = "ObsColumn"
    _uuid = "5c567168-0287-11ed-9d0f-e77f5194a415"
    _metakey = f"{_root}/uuid/{_uuid}/metastore"

    def read_file(
        self,
        filepath: Union[str, Path],
        satellite: Optional[str] = None,
        domain: Optional[str] = None,
        selection: Optional[str] = None,
        site: Optional[str] = None,
        species: Optional[str] = None,
        network: Optional[str] = None,
        instrument: Optional[str] = None,
        platform: str = "satellite",
        source_format: str = "openghg",
        if_exists: str = "auto",
        save_current: str = "auto",
        overwrite: bool = False,
        force: bool = False,
        compressor: Optional[Any] = None,
        filters: Optional[Any] = None,
        chunks: Optional[Dict] = None,
        optional_metadata: Optional[Dict] = None,
    ) -> dict:
        """Read column observation file

        Args:
            filepath: Path of observation file
            satellite: Name of satellite (if relevant)
            domain: For satellite only. If data has been selected on an area include the
                identifier name for domain covered. This can map to previously defined domains
                (see openghg_defs "domain_info.json" file) or a newly defined domain.
            selection: For satellite only, identifier for any data selection which has been
                performed on satellite data. This can be based on any form of filtering, binning etc.
                but should be unique compared to other selections made e.g. "land", "glint", "upperlimit".
                If not specified, domain will be used.
            site : Site code/name (if relevant). Can include satellite OR site.
            species: Species name or synonym e.g. "ch4"
            instrument: Instrument name e.g. "TANSO-FTS"
            network: Name of in-situ or satellite network e.g. "TCCON", "GOSAT"
            platform: Type of platform. Should be one of:
                - "satellite"
                - "site"
            source_format : Type of data being input e.g. openghg (internal format)
            if_exists: What to do if existing data is present.
                - "auto" - checks new and current data for timeseries overlap
                   - adds data if no overlap
                   - raises DataOverlapError if there is an overlap
                - "new" - just include new data and ignore previous
                - "combine" - replace and insert new data into current timeseries
            save_current: Whether to save data in current form and create a new version.
                - "auto" - this will depend on if_exists input ("auto" -> False), (other -> True)
                - "y" / "yes" - Save current data exactly as it exists as a separate (previous) version
                - "n" / "no" - Allow current data to updated / deleted
            overwrite: Deprecated. This will use options for if_exists="new".
            force: Force adding of data even if this is identical to data stored.
            compressor: A custom compressor to use. If None, this will default to
                `Blosc(cname="zstd", clevel=5, shuffle=Blosc.SHUFFLE)`.
                See https://zarr.readthedocs.io/en/stable/api/codecs.html for more information on compressors.
            filters: Filters to apply to the data on storage, this defaults to no filtering. See
                https://zarr.readthedocs.io/en/stable/tutorial.html#filters for more information on picking filters.
            chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
                for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
                See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking.
                To disable chunking pass in an empty dictionary.
            optional_metadata: Allows to pass in additional tags to distinguish added data. e.g {"project":"paris", "baseline":"Intem"}
        Returns:
            dict: Dictionary of datasource UUIDs data assigned to
        """
        from openghg.types import ColumnTypes
        from openghg.util import clean_string, load_column_parser, check_if_need_new_version, synonyms

        # TODO: Evaluate which inputs need cleaning (if any)
        satellite = clean_string(satellite)
        site = clean_string(site)
        species = clean_string(species)
        if species is not None:
            species = synonyms(species)
        domain = clean_string(domain)
        network = clean_string(network)
        instrument = clean_string(instrument)
        platform = clean_string(platform)

        if overwrite and if_exists == "auto":
            logger.warning(
                "Overwrite flag is deprecated in preference to `if_exists` (and `save_current`) inputs."
                "See documentation for details of these inputs and options."
            )
            if_exists = "new"

        # Making sure new version will be created by default if force keyword is included.
        if force and if_exists == "auto":
            if_exists = "new"

        new_version = check_if_need_new_version(if_exists, save_current)

        filepath = Path(filepath)

        try:
            source_format = ColumnTypes[source_format.upper()].value
        except KeyError:
            raise ValueError(f"Unknown data type {source_format} selected.")

        # Load the data retrieve object
        parser_fn = load_column_parser(source_format=source_format)

        _, unseen_hashes = self.check_hashes(filepaths=filepath, force=force)

        if not unseen_hashes:
            return {}

        filepath = next(iter(unseen_hashes.values()))

        if chunks is None:
            chunks = {}

        # Define parameters to pass to the parser function
        param = {
            "data_filepath": filepath,
            "satellite": satellite,
            "domain": domain,
            "selection": selection,
            "site": site,
            "species": species,
            "network": network,
            "instrument": instrument,
            "platform": platform,
            "chunks": chunks,
        }

        obs_data = parser_fn(**param)

        # TODO: Add in schema and checks for ObsColumn
        # # Checking against expected format for ObsColumn
        # for split_data in obs_data.values():
        #     col_data = split_data["data"]
        #     ObsColumn.validate_data(col_data)

        # TODO: Do we need to do include a split here of some kind, since
        # this could be "site" or "satellite" keys.
        # platform = list(obs_data.keys())[0]["metadata"]["platform"]

        required = ("satellite", "selection", "domain", "site", "species", "network")

        if optional_metadata:
            common_keys = set(required) & set(optional_metadata.keys())

            if common_keys:
                raise ValueError(
                    f"The following optional metadata keys are already present in required keys: {', '.join(common_keys)}"
                )
            else:
                for key, parsed_data in obs_data.items():
                    parsed_data["metadata"].update(optional_metadata)

        data_type = "column"
        datasource_uuids = self.assign_data(
            data=obs_data,
            if_exists=if_exists,
            new_version=new_version,
            data_type=data_type,
            required_keys=required,
            min_keys=3,
            compressor=compressor,
            filters=filters,
        )

        # TODO: MAY NEED TO ADD BACK IN OR CAN DELETE
        # update_keys = ["start_date", "end_date", "latest_version"]
        # obs_data = update_metadata(data_dict=obs_data, uuid_dict=datasource_uuids, update_keys=update_keys)

        # obs_store.add_datasources(
        #     uuids=datasource_uuids, data=obs_data, metastore=metastore, update_keys=update_keys
        # )

        # Record the file hash in case we see this file again
        self.store_hashes(unseen_hashes)

        return datasource_uuids

    # TODO: Add in transform method for gosat and tropomi raw data files
    #  - Included emissions version as starting point to be updated.
    # @staticmethod
    # def transform_data(
    #     datapath: Union[str, Path],
    #     database: str,
    #     overwrite: bool = False,
    #     **kwargs: Dict,
    # ) -> Dict:
    #     """
    #     Read and transform an emissions database. This will find the appropriate
    #     parser function to use for the database specified. The necessary inputs
    #     are determined by which database ie being used.

    #     The underlying parser functions will be of the form:
    #         - openghg.transform.emissions.parse_{database.lower()}
    #             - e.g. openghg.transform.emissions.parse_edgar()

    #     Args:
    #         datapath: Path to local copy of database archive (for now)
    #         database: Name of database
    #         overwrite: Should this data overwrite currently stored data
    #             which matches.
    #         **kwargs: Inputs for underlying parser function for the database.
    #             Necessary inputs will depend on the database being parsed.

    #     TODO: Could allow Callable[..., Dataset] type for a pre-defined function be passed
    #     """
    #     import inspect
    #     from openghg.store import assign_data, load_metastore, datasource_lookup
    #     from openghg.types import EmissionsDatabases
    #     from openghg.util import load_emissions_database_parser

    #     datapath = Path(datapath)

    #     try:
    #         data_type = EmissionsDatabases[database.upper()].value
    #     except KeyError:
    #         raise ValueError(f"Unable to transform '{database}' selected.")

    #     # Load the data retrieve object
    #     parser_fn = load_emissions_database_parser(database=database)

    #     em_store = Emissions.load()

    #     # Load in the metadata store
    #     metastore = load_metastore(key=em_store._metakey)

    #     # Find all parameters that can be accepted by parse function
    #     all_param = list(inspect.signature(parser_fn).parameters.keys())

    #     # Define parameters to pass to the parser function from kwargs
    #     param: Dict[Any, Any] = {key: value for key, value in kwargs.items() if key in all_param}
    #     param["datapath"] = datapath  # Add datapath explicitly (for now)

    #     emissions_data = parser_fn(**param)

    #     # Checking against expected format for Emissions
    #     for split_data in emissions_data.values():
    #         em_data = split_data["data"]
    #         Emissions.validate_data(em_data)

    #     required = ("species", "source", "domain", "date")
    #     lookup_results = datasource_lookup(metastore=metastore, data=emissions_data, required_keys=required)

    #     data_type = "emissions"
    #     overwrite = False
    #     datasource_uuids = assign_data(
    #         data_dict=emissions_data,
    #         lookup_results=lookup_results,
    #         overwrite=overwrite,
    #         data_type=data_type,
    #     )

    #     em_store.add_datasources(uuids=datasource_uuids, data=emissions_data, metastore=metastore)

    #     em_store.save()
    #     metastore.close()

    #     return datasource_uuids

    # TODO: Define and add schema methods for ObsColumn
    # @staticmethod
    # def schema(species: str, platform: str = "satellite") -> DataSchema:
    #     """
    #     Define schema for emissions Dataset.

    #     Includes column data for each time point:
    #         - standardised species and column name (e.g. "xch4")
    #         - expected dimensions: ("time")

    #     Expected data types for all variables and coordinates also included.

    #     Returns:
    #         DataSchema : Contains schema for Emissions.
    #     """
    #     data_vars: Dict[str, Tuple[str, ...]] \
    #         = {"flux": ("time", "lat", "lon")}
    #     dtypes = {"lat": np.floating,
    #               "lon": np.floating,
    #               "time": np.datetime64,
    #               "flux": np.floating}

    #     data_format = DataSchema(data_vars=data_vars,
    #                              dtypes=dtypes)

    #     return data_format

    # @staticmethod
    # def validate_data(data: Dataset) -> None:
    #     """
    #     Validate input data against Emissions schema - definition from
    #     Emissions.schema() method.

    #     Args:
    #         data : xarray Dataset in expected format

    #     Returns:
    #         None

    #         Raises a ValueError with details if the input data does not adhere
    #         to the Emissions schema.
    #     """
    #     data_schema = Emissions.schema()
    #     data_schema.validate_data(data)



from __future__ import annotations
import logging
from typing import List, Optional, Union, Dict, Tuple
import plotly.graph_objects as go
import numpy as np
import base64
from typing import TYPE_CHECKING

from openghg.util import get_species_info, synonyms, get_datapath

if TYPE_CHECKING:
    from openghg.dataobjects import ObsData


logger = logging.getLogger("openghg.plotting")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


def _latex2html(latex_string: str) -> str:
    """Replace latex sub/superscript formatting with html.
    Written because the latex formatting in Plotly seems inconsistent
    (works in Notebooks, but not VSCode at the moment).

    Args:
        latex_string: String containing LaTeX math mode (including $$)
    Returns:
        str: string with matched sub-strings replaced with equivalent html.
    """

    replacements = {
        "$^2$": "<sup>2</sup>",
        "$^{-1}$": "<sup>-1</sup>",
        "$^{-2}$": "<sup>-2</sup>",
        "$_2$": "<sub>2</sub>",
        "$_3$": "<sub>3</sub>",
        "$_4$": "<sub>4</sub>",
        "$_5$": "<sub>5</sub>",
        "$_6$": "<sub>6</sub>",
    }

    html_string = latex_string
    for rep in replacements:
        html_string = html_string.replace(rep, replacements[rep])

    return html_string


def _plot_remove_gaps(
    x_data: np.ndarray, y_data: np.ndarray, gap: Optional[int] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Insert NaNs between big gaps in the data.
    Prevents connecting lines being drawn

    Args:
        x_data: plot timeseries (numpy timestamp)
        y_data: data array
        gap: gap beyond which a NaN is introducted (nanoseconds, defaults to 1 day)
    Returns:
        x, y: x and y arrays to plot
    """
    if gap is None:
        # ns in a day
        gap = 24 * 60 * 60 * 1000000000

    gap_idx = np.where(np.diff(x_data.astype(int)) > gap)[0]
    x_data_plot = np.insert(x_data, gap_idx + 1, values=x_data[0])
    y_data_plot = np.insert(y_data, gap_idx + 1, values=np.nan)

    return x_data_plot, y_data_plot


def _plot_legend_position(ascending: bool) -> Tuple[Dict, Dict]:
    """Position of legend and logo,
    depending on whether data is ascending or descending

    Args:
        ascending: Is the data ascending
    Returns:
        Dict, Dict: Plotly legend and logo position parameters
    """
    if ascending:
        legend_pos = {"yanchor": "top", "xanchor": "left", "y": 0.99, "x": 0.01}
        logo_pos = {"yanchor": "bottom", "xanchor": "right", "y": 0.01, "x": 0.99}
    else:
        legend_pos = {"yanchor": "top", "xanchor": "right", "y": 0.99, "x": 0.99}
        logo_pos = {"yanchor": "bottom", "xanchor": "left", "y": 0.01, "x": 0.01}

    return legend_pos, logo_pos


def _plot_logo(
    logo_pos: Dict,
) -> Dict:
    """Create Plotly dictionary for logo

    Args:
        logo_pos: Dictionary containing the position of the logo
    Returns:
        dict: Dictionary containing logo + position parameters
    """
    logo_bytes = get_datapath("OpenGHG_Logo_NoText_transparent_200x200.png").read_bytes()
    logo = base64.b64encode(logo_bytes)

    logo_dict = dict(
        source="data:image/png;base64,{}".format(logo.decode()),
        xref="x domain",
        yref="y domain",
        sizex=0.1,
        sizey=0.1,
    )
    logo_dict.update(logo_pos)

    return logo_dict


def plot_timeseries(
    data: Union[ObsData, List[ObsData]],
    xvar: Optional[str] = None,
    yvar: Optional[str] = None,
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    units: Optional[str] = None,
    logo: Optional[bool] = True,
) -> go.Figure:
    """Plot a timeseries

    Args:
        data: ObsData object or list of objects
        xvar: x axis variable, defaults to time
        yvar: y axis variable, defaults to species
        title: Title for figure
        xlabel: Label for x axis
        ylabel: Label for y axis
        units: Units for y axis
        logo: Show the OpenGHG logo
    Returns:
        go.Figure: Plotly Graph Object Figure
    """
    from openghg.util import load_internal_json

    if not data:
        logger.warning("No data to plot, returning")
        return None

    if not isinstance(data, list):
        data = [data]

    # Get species info
    species_info = get_species_info()

    # Get some general attributes
    attributes_data = load_internal_json("attributes.json")

    font = {"size": 14}

    margin = {"l": 20, "r": 20, "t": 20, "b": 20}

    if title is not None:
        title_layout = {"text": title, "y": 0.9, "x": 0.5, "xanchor": "center", "yanchor": "top"}
        layout = go.Layout(
            title=title_layout, xaxis=dict(title=xlabel), yaxis=dict(title=ylabel), font=font, margin=margin
        )
    else:
        layout = go.Layout(font=font, margin=margin)

    # Create a single figure
    fig = go.Figure(layout=layout)

    species_strings = []
    unit_strings = []

    # Loop through inlets/species
    for i, to_plot in enumerate(data):
        metadata = to_plot.metadata
        dataset = to_plot.data

        species = metadata["species"]
        site = metadata["site"]
        inlet = metadata["inlet"]

        species_string = _latex2html(species_info[synonyms(species, lower=False)]["print_string"])

        legend_text = f"{species_string} - {site.upper()} ({inlet})"

        if xvar is not None:
            x_data = dataset[xvar]
        else:
            x_data = dataset.time

        if yvar is not None:
            y_data = dataset[yvar]
        else:
            try:
                y_data = dataset[species]
            except KeyError:
                y_data = dataset["mf"]

        if units is not None or len(data) > 0:
            data_attrs = y_data.attrs
            data_units = data_attrs.get("units", "1")

            if i == 0:
                if units:
                    unit_interpret = attributes_data["unit_interpret"]
                    unit_value = unit_interpret.get(units, "1")
                else:
                    unit_value = data_units

            unit_conversion = float(data_units) / float(unit_value)
        else:
            unit_conversion = 1
            # TODO: Not sure what is expected for unit_value here
            unit_value = "1"

        y_data *= unit_conversion

        unit_string = attributes_data["unit_print"][unit_value]

        # Add NaNs where there are large data gaps
        x_data_plot, y_data_plot = _plot_remove_gaps(x_data.values, y_data.values)

        # Convert unit string to html
        unit_string_html = _latex2html(unit_string)

        # Create plot
        fig.add_trace(
            go.Scatter(
                name=legend_text,
                x=x_data_plot,
                y=y_data_plot,
                mode="lines",
                hovertemplate="%{x|%Y-%m-%d %H:%M}<br> %{y:.1f} " + unit_string_html,
            )
        )

        # Save units and species names for axis labels
        unit_strings.append(unit_string_html)
        species_strings.append(species_string)

    # Determine whether data is ascending or descending (positioning of legend)
    y_data_diff = y_data.diff(dim="time").mean()
    if y_data_diff >= 0:
        ascending = True
    else:
        ascending = False

    if len(set(unit_strings)) > 1:
        raise NotImplementedError("Can't plot two different units yet")

    # Write species and units on y-axis
    if ylabel is not None:
        fig.update_yaxes(title=ylabel)
    else:
        ytitle = ", ".join(set(species_strings)) + " (" + unit_strings[0] + ")"
        fig.update_yaxes(title=ytitle)

    if xlabel is None:
        xlabel = "Date"

    fig.update_xaxes(title=xlabel)

    # Position the legend
    legend_pos, logo_pos = _plot_legend_position(ascending)
    fig.update_layout(legend=legend_pos, template="seaborn")

    # Add OpenGHG logo
    if logo:
        logo_dict = _plot_logo(logo_pos)
        fig.add_layout_image(logo_dict)

    return fig



from dataclasses import dataclass
from typing import Dict, List


@dataclass(frozen=True)
class ChunkingSchema:
    """Chunking scheme for zarr storage.

    Args:
        variable: Name of the variable
        chunks: Dictionary of chunk sizes for each dimension
        secondary_dimensions: List of secondary dimensions to chunk over
    """

    variable: str
    chunks: Dict[str, int]
    secondary_dims: List[str]
    # max chunk size in megabytes
    max_chunk_size: int = 300



from pathlib import Path
from typing import Dict, List, MutableMapping, Optional, Union, cast

import xarray as xr


def parse_openghg(
    data_filepath: Union[str, Path],
    satellite: Optional[str] = None,
    domain: Optional[str] = None,
    selection: Optional[str] = None,
    site: Optional[str] = None,
    species: Optional[str] = None,
    network: Optional[str] = None,
    instrument: Optional[str] = None,
    platform: str = "satellite",
    chunks: Optional[Dict] = None,
    **kwargs: str,
) -> Dict:
    """
    Parse and extract data from pre-formatted netcdf file which already
    matches expected OpenGHG format.

    The arguments specified below are the metadata needed to store this
    surface observation file within the object store. If these keywords are
    not included within the attributes of the netcdf file being passed then
    these arguments must be specified.

    For column data this can either be a satellite (e.g. satellite="GOSAT") or a
    site (site="RUN", network="TCCON"). Either can be specified or this function
    will attempt to extract this from the data file.

    Args:
        data_filepath: Path of observation file
        satellite: Name of satellite (if relevant)
        domain: For satellite only. If data has been selected on an area include the
            identifier name for domain covered. This can map to previously defined domains
            (see openghg_defs "domain_info.json" file) or a newly defined domain.
        selection: For satellite only, identifier for any data selection which has been
            performed on satellite data. This can be based on any form of filtering, binning etc.
            but should be unique compared to other selections made e.g. "land", "glint", "upperlimit".
            If not specified, domain will be used.
        site : Site code/name (if relevant). Can include satellite OR site.
        species: Species name or synonym e.g. "ch4"
        instrument: Instrument name e.g. "TANSO-FTS"
        network: Name of in-situ or satellite network e.g. "TCCON", "GOSAT"
        platform: Type of platform. Should be one of:
            - "satellite"
            - "site"
            Note: this will be superceded if site or satellite keywords are specified.
        chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
            for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
            See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking.
            To disable chunking pass in an empty dictionary.
        kwargs: Any additional attributes to be associated with the data.
    Returns:
        Dict : Dictionary of source_name : data, metadata, attributes
    """
    from openghg.standardise.meta import define_species_label
    from openghg.util import clean_string

    # from openghg.standardise.meta import metadata_default_keys, assign_attributes

    data_filepath = Path(data_filepath)

    if data_filepath.suffix.lower() != ".nc":
        raise ValueError("Input file must be a .nc (netcdf) file.")

    data = xr.open_dataset(data_filepath).chunk(chunks)

    # Extract current attributes from input data
    attributes = cast(MutableMapping, data.attrs)

    if satellite is not None or platform == "satellite":
        metadata_required = metadata_default_satellite_column()
        metadata_required.remove("selection")
        platform = "satellite"
    elif site is not None or platform == "site":
        metadata_required = metadata_default_site_column()
        platform = "site"

    if platform == "satellite":
        if domain is None:
            raise ValueError(
                "For satellite data, please specify selected domain."
                "This can be 'global' if no selection has been made."
            )

    # Define metadata based on input arguments.
    metadata_initial = {
        "site": site,
        "satellite": satellite,
        "instrument": instrument,
        "species": species,
        "domain": domain,
        "network": network,
        "platform": platform,
        "data_type": "column",
        "source_format": "openghg",
    }

    # TODO: Tidy this up a bit (maybe split some into a separate function?)
    # and incorporate kwargs

    metadata = {}
    key_translation = satellite_attribute_translation()
    # Populate metadata with values from attributes if inputs have not been passed
    for key, value in metadata_initial.items():
        if key in metadata_required:
            # Extract equivalent key from passed file if present using translation
            try:
                attr_keys = key_translation[key]
            except KeyError:
                attr_keys = key

            # Make sure this is a list for cases with multiple options
            if isinstance(attr_keys, str):
                attr_keys = [attr_keys]

            if value is None:
                # Extract value from attributes if this has not been specified
                for attr_key in attr_keys:
                    try:
                        metadata[key] = attributes[attr_key]
                    except KeyError:
                        continue
                    else:
                        break
                else:
                    raise ValueError(f"Input '{key}' must be specified if not included in data attributes.")
            else:
                # If attributes are present, check these match to inputs passed
                for attr_key in attr_keys:
                    if attr_key in key_translation and attr_key in attributes:
                        attributes_value = attributes[attr_key]
                        if value != attributes_value:
                            # If inputs do not match attribute values, raise a ValueError
                            raise ValueError(
                                f"Input for '{key}': {value} does not match value in file attributes: {attributes_value}"
                            )
                metadata[key] = value

    # metadata = cast(Dict[str, str], metadata_initial)

    if selection is not None:
        metadata["selection"] = selection
    elif platform == "satellite":
        metadata["selection"] = domain

    # TODO: Add loose domain checking? If known domain is specified, make sure points
    # are within this for example.

    species = define_species_label(metadata["species"])[0]
    metadata["species"] = species

    if "site" in metadata:
        metadata["site"] = clean_string(metadata["site"])

    # Add data type to metadata
    metadata["data_type"] = "column"

    # Define remaining keys needed for metadata
    metadata_needed = [param for param in metadata_required if param not in metadata]

    for key in metadata_needed:
        try:
            attr_keys = key_translation[key]
        except KeyError:
            attr_keys = key

        if isinstance(attr_keys, str):
            attr_keys = [attr_keys]

        for attr_key in attr_keys:
            for attr_key in attr_keys:
                try:
                    metadata[key] = attributes[attr_key]
                except KeyError:
                    continue
                else:
                    break
            else:
                raise ValueError(f"Input '{key}' must be specified if not included in data attributes.")

    # In GOSAT UoL files (and copied into our files)
    #  - platform = "GOSAT"; sensor = "TANSO-FTS"
    # In TROPOMI S5P_OFFL_...nc files
    #  - platform = 'S5P'; sensor = "TROPOMI"
    # Could "platform" --> "satellite" perhaps?

    # site = metadata["site"]
    # network = metadata["network"]
    # species = metadata["species"]

    # # Allow site and network data to be treated in a case insensitive way
    # site_case_options = [site, site.upper(), site.lower()]
    # network_case_options = [network, network.upper(), network.lower()]

    # # Extract centralised data for site (if present)
    # site_data = load_json(filename="site_info.json")
    # for site_value in site_case_options:
    #     if site_value in site_data:
    #         site_info_all = site_data[site_value]
    #         break
    # else:
    #     print("Unknown site. Will attempt to extract metadata from dataset attributes or input keywords")
    #     site_info_all = {}

    # for network_value in network_case_options:
    #     if network_value in site_info_all:
    #         site_info = site_info_all[network_value]
    #         break
    # else:
    #     print(
    #         "Network {network} does not match with site {site}. Will attempt to extract metadata from dataset attributes or input keywords"
    #     )
    #     site_info = {}

    # if site_info:
    #     # Ensure keywords match to metadata names for station values
    #     # e.g. "station_longitude" derived from "longitude"
    #     for key in metadata_needed:
    #         prefix = "station_"
    #         if key.startswith(prefix):
    #             split_key = key.split("_")[1:]
    #             short_key_option1 = "_".join(split_key)

    #             split_key.insert(1, "station")  # to catch "height_station_masl"
    #             short_key_option2 = "_".join(split_key)

    #             short_key_options = [short_key_option1, short_key_option2]
    #             for short_key in short_key_options:
    #                 if short_key in site_info:
    #                     site_info[key] = site_info[short_key]
    #                     break

    # # Load attributes data for network if present
    # param_data = load_json(filename="attributes.json")
    # for network_value in network_case_options:
    #     if network_value in param_data:
    #         network_params = param_data[network_value]
    #         site_attributes = network_params["global_attributes"]
    #         break
    # else:
    #     site_attributes = {}

    # # Define sources of attributes to use when defining metadata
    # # The order here influences the hierarchy if keys appear multiple times.
    # # kwargs allow additional variables such as "station_longitude" to be included if needed.
    # attribute_sources = [attributes, kwargs, site_info, site_attributes]

    # # Search attributes sources (in order) and populate metadata
    # for param in metadata_needed:
    #     for source in attribute_sources:
    #         if param in source:
    #             metadata[param] = source[param]
    #             break
    #     else:
    #         raise ValueError(
    #             f"Cannot extract or infer '{param}' parameter needed for metadata from stored data, attributes or keywords"
    #         )

    # Update attributes to match metadata after cleaning
    attributes.update(metadata)

    # TODO: Decide if the key here should be more descriptive that just `species`
    gas_data = {species: {"metadata": metadata, "data": data, "attributes": attributes}}

    # gas_data = assign_attributes(data=gas_data, site=site, network=network)

    return gas_data


def metadata_default_satellite_column() -> List[str]:
    """
    Define default keys for satellite column data
    """
    default_keys = [
        "satellite",
        "species",
        "network",
        "instrument",
        "platform",
        "domain",
        "selection",
        "data_owner",
        "data_owner_email",
    ]

    return default_keys


def metadata_default_site_column() -> List[str]:
    """
    Define default keys for site column data
    """
    default_keys = [
        "site",
        "species",
        "network",
        "instrument",
        "platform",
        "data_owner",
        "data_owner_email",
    ]

    return default_keys


TranslationDict = Dict[str, Union[str, List[str]]]


def satellite_attribute_translation() -> TranslationDict:
    """
    Define translation between openghg keyword and input files.
    Currently includes:
     - GOSAT (University of Leicester product)
     - TROPOMI
    """
    # Current values within (at least) GOSAT, TROPOMI files
    # - values can contain lists as well as single string values
    keywords: TranslationDict = {
        "instrument": "sensor",
        "satellite": "platform",
        "network": "platform",
        "data_owner": "creator_name",
        "data_owner_email": "creator_email",
    }

    return keywords



from ._boundary_conditions import BoundaryConditions
from ._data_schema import DataSchema
from ._flux import Flux
from ._eulerian_model import EulerianModel
from ._footprints import Footprints
from ._infer_time import infer_date_range, update_zero_dim
from ._obsmobile import ObsMobile
from ._obscolumn import ObsColumn
from ._obssurface import ObsSurface
from ._populate import add_noaa_obspack
from ._metstore import METStore
from ._meta import data_class_info, get_data_class



from pathlib import Path
from typing import Any, Dict
import time
import zarr
import numcodecs
import itertools
import tempfile
from numcodecs import Blosc
import xarray as xr

from openghg.store.storage import get_zarr_encoding


def compare_compression(filepath: Path) -> Dict:
    """This compresses the given dataset to a temporary zarr ZipStore
    to test the compression ratio. A dictionary containing the final size of
    the data, the compression ratio and the time taken for the compression
    is returned"

    Args:
        filepath: Path to NetCDF file
    Returns:
        dict: Dictionary of compression information
    """
    codecs = numcodecs.blosc.list_compressors()
    compression_levels = [1, 3, 5]
    # These are just ints
    shuffles = [Blosc.NOSHUFFLE, Blosc.SHUFFLE, Blosc.BITSHUFFLE]
    codec_combinations = sorted(set(list(itertools.product(codecs, compression_levels, shuffles))))

    filepath = Path(filepath)
    orig_size = filepath.stat().st_size
    compression_info: Dict[str, Any] = {"original_filesize": orig_size}
    print(f"Original file size: {orig_size}")

    best_ratio = 0.0
    shortest_time = 1.0e9
    quickest_codec = ""
    best_compression_codec = ""

    with tempfile.TemporaryDirectory() as tmpdir:
        for codec, comp_level, shuffle in codec_combinations:
            with xr.open_dataset(filepath) as ds:
                codec_str = f"blosc_codec_{codec}_complevel_{comp_level}_shuffle_{shuffle}"

                zarr_path = Path(tmpdir, f"zarr_store_{codec_str}.zip")
                store = zarr.storage.ZipStore(zarr_path)

                compressor = Blosc(cname=codec, clevel=comp_level, shuffle=shuffle)
                encoding = get_zarr_encoding(data_vars=ds.data_vars, compressor=compressor)

                start = time.perf_counter()
                ds.to_zarr(store=store, encoding=encoding)
                time_taken = time.perf_counter() - start

                compressed_size = zarr_path.stat().st_size

                ratio = orig_size / compressed_size
                print(f"Finished {codec_str}")

                compression_info[codec_str] = {
                    "ratio": ratio,
                    "compressed_size_bytes": compressed_size,
                    "time_taken": time_taken,
                }

                if ratio > best_ratio:
                    best_ratio = ratio
                    best_compression_codec = codec_str

                if time_taken < shortest_time:
                    shortest_time = time_taken
                    quickest_codec = codec_str

    compression_info["best_compression_codec"] = best_compression_codec
    compression_info["quickest_codec"] = quickest_codec

    return compression_info



from typing import Any, Dict, List, Optional, Union
from openghg.dataobjects import ObsData
from openghg.objectstore import get_writable_bucket
from openghg.util import running_on_hub, load_json
import openghg_defs
import logging

logger = logging.getLogger("openghg.retrieve")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


def retrieve_atmospheric(
    site: str,
    species: Optional[Union[str, List]] = None,
    inlet: Optional[str] = None,
    sampling_height: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    force_retrieval: bool = False,
    data_level: int = 2,
    dataset_source: Optional[str] = None,
    store: Optional[str] = None,
    update_mismatch: str = "never",
    force: bool = False,
) -> Union[ObsData, List[ObsData], None]:
    """Retrieve ICOS atmospheric measurement data. If data is found in the object store it is returned. Otherwise
    data will be retrieved from the ICOS Carbon Portal. Data retrieval from the Carbon Portal may take a short time.
    If only a single data source is found an ObsData object is returned, if multiple a list of ObsData objects
    if returned, if nothing then None.

    Args:
        site: Site code
        species: Species name
        inlet: Height of the inlet for sampling in metres.
        sampling_height: Alias for inlet
        start_date: Start date
        end_date: End date
        force_retrieval: Force the retrieval of data from the ICOS Carbon Portal
        data_level: ICOS data level (1, 2)
        - Data level 1: Near Real Time Data (NRT) or Internal Work data (IW).
        - Data level 2: The final quality checked ICOS RI data set, published by the CFs,
                        to be distributed through the Carbon Portal.
                        This level is the ICOS-data product and free available for users.
        See https://icos-carbon-portal.github.io/pylib/modules/#stationdatalevelnone
        dataset_source: Dataset source name, for example ICOS, InGOS, European ObsPack
        store: Name of object to search/store data to
        update_mismatch: This determines how mismatches between the "metadata" derived from
            stored data and "attributes" derived from ICOS Header are handled.
            This includes the options:
                - "never" - don't update mismatches and raise an AttrMismatchError
                - "from_source" / "attributes" - update mismatches based on attributes from ICOS Header
                - "from_definition" / "metadata" - update mismatches based on input metadata
        force: Force adding of data even if this is identical to data stored (checked based on previously retrieved file hashes).
    Returns:
        ObsData, list[ObsData] or None
    """
    return retrieve(
        site=site,
        species=species,
        inlet=inlet,
        sampling_height=sampling_height,
        start_date=start_date,
        end_date=end_date,
        force_retrieval=force_retrieval,
        data_level=data_level,
        dataset_source=dataset_source,
        update_mismatch=update_mismatch,
        store=store,
        force=force,
    )


def retrieve(**kwargs: Any) -> Union[ObsData, List[ObsData], None]:
    """Retrieve data from the ICOS Carbon Portal. If data is found in the local object store
    it will be retrieved from there first.

    This function detects the running environment and routes the call
    to either the cloud or local search function.

    Example / commonly used arguments are given below.

    Args:
        site: Site code
        species: Species name
        start_date: Start date
        end_date: End date
        inlet: Height of the inlet for sampling in metres.
        sampling_height: Alias for inlet
        force_retrieval: Force the retrieval of data from the ICOS Carbon Portal
        data_level: ICOS data level (1, 2)
        - Data level 1: Near Real Time Data (NRT) or Internal Work data (IW).
        - Data level 2: The final quality checked ICOS RI data set, published by the CFs,
                        to be distributed through the Carbon Portal.
                        This level is the ICOS-data product and free available for users.
        See https://icos-carbon-portal.github.io/pylib/modules/#stationdatalevelnone
        update_mismatch: This determines how mismatches between the "metadata" derived from
            stored data and "attributes" derived from ICOS Header are handled.
            This includes the options:
                - "never" - don't update mismatches and raise an AttrMismatchError
                - "from_source" / "attributes" - update mismatches based on attributes from ICOS Header
                - "from_definition" / "metadata" - update mismatches based on input metadata
    Returns:
        ObsData, list[ObsData] or None
    """
    from io import BytesIO
    from xarray import load_dataset
    from openghg.cloud import call_function, unpackage

    # The hub is the only place we want to make remote calls
    if running_on_hub():
        raise NotImplementedError("Cloud functionality marked for rewrite.")
        post_data: Dict[str, Union[str, Dict]] = {}
        post_data["function"] = "retrieve_icos"
        post_data["search_terms"] = kwargs

        call_result = call_function(data=post_data)

        content = call_result["content"]
        found = content["found"]

        if not found:
            return None

        observations = content["data"]

        obs_data = []
        for package in observations.values():
            unpackaged = unpackage(data=package)
            buf = BytesIO(unpackaged["data"])
            ds = load_dataset(buf)
            obs = ObsData(data=ds, metadata=unpackaged["metadata"])

            obs_data.append(obs)

        if len(obs_data) == 1:
            return obs_data[0]
        else:
            return obs_data
    else:
        return local_retrieve(**kwargs)


def local_retrieve(
    site: str,
    species: Optional[Union[str, List]] = None,
    inlet: Optional[str] = None,
    sampling_height: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    force_retrieval: bool = False,
    data_level: int = 2,
    dataset_source: Optional[str] = None,
    store: Optional[str] = None,
    update_mismatch: str = "never",
    force: bool = False,
    **kwargs: Any,
) -> Union[ObsData, List[ObsData], None]:
    """Retrieve ICOS atmospheric measurement data. If data is found in the object store it is returned. Otherwise
    data will be retrieved from the ICOS Carbon Portal. Data retrieval from the Carbon Portal may take a short time.
    If only a single data source is found an ObsData object is returned, if multiple a list of ObsData objects
    if returned, if nothing then None.

    Args:
        site: Site code
        species: Species name
        inlet: Height of the inlet for sampling in metres.
        sampling_height: Alias for inlet
        start_date: Start date
        end_date: End date
        force_retrieval: Force the retrieval of data from the ICOS Carbon Portal
        data_level: ICOS data level (1, 2)
        - Data level 1: Near Real Time Data (NRT) or Internal Work data (IW).
        - Data level 2: The final quality checked ICOS RI data set, published by the CFs,
                        to be distributed through the Carbon Portal.
                        This level is the ICOS-data product and free available for users.
        See https://icos-carbon-portal.github.io/pylib/modules/#stationdatalevelnone
        dataset_source: Dataset source name, for example ICOS, InGOS, European ObsPack
        store: Name of object to search/store data to
        update_mismatch: This determines how mismatches between the "metadata" derived from
            stored data and "attributes" derived from ICOS Header are handled.
            This includes the options:
                - "never" - don't update mismatches and raise an AttrMismatchError
                - "from_source" / "attributes" - update mismatches based on attributes from ICOS Header
                - "from_definition" / "metadata" - update mismatches based on input metadata
        force: Force adding of data even if this is identical to data stored (checked based on previously retrieved file hashes).
    Returns:
        ObsData, list[ObsData] or None
    """
    from openghg.retrieve import search_surface
    from openghg.store import ObsSurface
    from openghg.util import to_lowercase

    if not 1 <= data_level <= 2:
        logger.error("Error: data level must be 1 or 2.")

    if sampling_height and inlet is None:
        inlet = sampling_height
    elif sampling_height and inlet:
        logger.warning(f"Both sampling height and inlet specified. Using inlet value of {inlet}")

    # NOTE - we skip ranking here, will we be ranking ICOS data?
    results = search_surface(
        site=site,
        species=species,
        inlet=inlet,
        network="ICOS",
        data_source="icoscp",
        start_date=start_date,
        end_date=end_date,
        icos_data_level=data_level,
        dataset_source=dataset_source,
        store=store,
    )

    if results and not force_retrieval:
        obs_data = results.retrieve_all()
    else:
        # We'll also need to check we have current data
        standardised_data = _retrieve_remote(
            site=site,
            species=species,
            data_level=data_level,
            dataset_source=dataset_source,
            inlet=inlet,
            sampling_height=sampling_height,
            update_mismatch=update_mismatch,
        )

        if standardised_data is None:
            return None

        bucket = get_writable_bucket(name=store)
        with ObsSurface(bucket=bucket) as obs:
            obs.store_data(data=standardised_data, force=force)

        # Create the expected ObsData type
        obs_data = []
        for data in standardised_data.values():
            measurement_data = data["data"]
            # These contain URLs that are case sensitive so skip lowercasing these
            skip_keys = [
                "citation_string",
                "instrument_data",
                "dobj_pid",
                "dataset_source",
            ]
            metadata = to_lowercase(data["metadata"], skip_keys=skip_keys)
            obs_data.append(ObsData(data=measurement_data, metadata=metadata))

    if isinstance(obs_data, list) and len(obs_data) == 1:
        return obs_data[0]
    else:
        return obs_data


def _retrieve_remote(
    site: str,
    data_level: int,
    species: Optional[Union[str, List]] = None,
    inlet: Optional[str] = None,
    sampling_height: Optional[str] = None,
    dataset_source: Optional[str] = None,
    update_mismatch: str = "never",
) -> Optional[Dict]:
    """Retrieve ICOS data from the ICOS Carbon Portal and standardise it into
    a format expected by OpenGHG. A dictionary of metadata and Datasets

    Args:
        site: ICOS site code, for site codes see
        https://www.icos-cp.eu/observations/atmosphere/stations
        data_level: ICOS data level (1, 2)
        - Data level 1: Near Real Time Data (NRT) or Internal Work data (IW).
        - Data level 2: The final quality checked ICOS RI data set, published by the CFs,
                        to be distributed through the Carbon Portal.
                        This level is the ICOS-data product and free available for users.
        See https://icos-carbon-portal.github.io/pylib/modules/#stationdatalevelnone
        species: Species name
        inlet: Height of the inlet for sampling in metres.
        sampling_height: Alias for inlet
        dataset_source: Dataset source name, for example ICOS, InGOS, European ObsPack
        update_mismatch: This determines how mismatches between the "metadata" derived from
            stored data and "attributes" derived from ICOS Header are handled.
            This includes the options:
                - "never" - don't update mismatches and raise an AttrMismatchError
                - "from_source" / "attributes" - update mismatches based on attributes from ICOS Header
                - "from_definition" / "metadata" - update mismatches based on input metadata
    Returns:
        dict or None: Dictionary of processed data and metadata if found
    """
    # icoscp isn't available to conda so we've got to resort to this for now
    try:
        from icoscp.cpb.dobj import Dobj  # type: ignore
        from icoscp.station import station  # type: ignore
    except ImportError:
        raise ImportError(
            "Cannot import icoscp, if you've installed OpenGHG using conda please run: pip install icoscp"
        )

    import re
    from openghg.standardise.meta import assign_attributes
    from openghg.util import format_inlet
    from pandas import to_datetime

    if species is None:
        species = ["CO", "CO2", "CH4"]

    if not isinstance(species, list):
        species = [species]

    if sampling_height and inlet is None:
        inlet = sampling_height
    elif sampling_height and inlet:
        logger.warning(f"Both sampling height and inlet specified. Using inlet value of {inlet}")

    # We should first check if it's stored in the object store
    # Will need to make sure ObsSurface can accept the datasets we
    # create from the ICOS data
    stat = station.get(stationId=site.upper())

    if not stat.valid:
        logger.error("Please check you have passed a valid ICOS site and have a working internet connection.")
        return None

    data_pids = stat.data(level=data_level)

    # We want to get the PIDs of the data for each species here
    species_upper = [s.upper() for s in species]
    # For this see https://stackoverflow.com/a/55335207
    search_str = r"\b(?:{})\b".format("|".join(map(re.escape, species_upper)))
    # Now filter the dataframe so we can extract the PIDS
    # We filter out any data that contains "Obspack" or "csv" in the specLabel
    # Also filter out some drought files which cause trouble being read in
    # For some reason they have separate station record pages that contain "ATMO_"
    filtered_sources = data_pids[
        data_pids["specLabel"].str.contains(search_str)
        & ~data_pids["specLabel"].str.contains("Obspack")
        & ~data_pids["specLabel"].str.contains("csv")
        & ~data_pids["station"].str.contains("ATMO_")
    ]

    if filtered_sources.empty:
        species_lower = [s.lower() for s in species]
        # For this see https://stackoverflow.com/a/55335207
        search_str = r"\b(?:{})\b".format("|".join(map(re.escape, species_lower)))
        # Now filter the dataframe so we can extract the PIDS
        filtered_sources = data_pids[
            data_pids["specLabel"].str.contains(search_str)
            & ~data_pids["specLabel"].str.contains("Obspack")
            & ~data_pids["specLabel"].str.contains("csv")
        ]

    if inlet is not None:
        inlet = str(float(inlet.rstrip("m")))
        height_filter = [inlet in str(x) for x in filtered_sources["samplingheight"]]
        filtered_sources = filtered_sources[height_filter]

    if filtered_sources.empty:
        logger.error(
            f"No sources found for {species} at {site}. Please check with the ICOS Carbon Portal that this data is available."
        )
        return None

    # Now extract the PIDs along with some data about them
    dobj_urls = filtered_sources["dobj"].tolist()

    # Load our site metadata for a few things like the station's long_name that
    # isn't in the ICOS metadata in the way we want it at the momenet - 2023-03-20
    site_info_fpath = openghg_defs.site_info_file
    openghg_site_metadata = load_json(path=site_info_fpath)

    standardised_data: Dict[str, Dict] = {}

    for n, dobj_url in enumerate(dobj_urls):
        dobj = Dobj(dobj_url)
        logger.info(f"Retrieving {dobj_url}...")
        # We've got to jump through some hoops here to try and avoid the NOAA
        # ObsPack GlobalView data
        try:
            if "globalview" in dobj.meta["references"]["doi"]["titles"][0]["title"].lower():
                logger.info(f"Skipping {dobj_url} as ObsPack GlobalView detected.")
                continue
        except KeyError:
            pass

        try:
            dobj_dataset_source = dobj.meta["specification"]["project"]["self"]["label"]
        except KeyError:
            dobj_dataset_source = "NA"
            logger.warning("Unable to read project information from dobj.")

        if dataset_source is not None and dataset_source.lower() != dobj_dataset_source.lower():
            continue

        # We need to pull the data down as .info (metadata) is populated further on this step
        dataframe = dobj.get()
        # This is the metadata, dobj.info and dobj.meta are equal
        dobj_info = dobj.meta

        attributes = {}

        specific_info = dobj_info["specificInfo"]
        col_data = specific_info["columns"]

        # Get the species this dobj holds information for
        not_the_species = {"TIMESTAMP", "Flag", "NbPoints", "Stdev"}
        species_info = next(i for i in col_data if i["label"] not in not_the_species)

        measurement_type = species_info["valueType"]["self"]["label"].lower()
        units = species_info["valueType"]["unit"].lower()
        the_species = species_info["label"]

        species_info = next(item for item in col_data if str(item["label"]).lower() == the_species.lower())

        attributes["species"] = the_species
        acq_data = specific_info["acquisition"]
        station_data = acq_data["station"]

        to_store: Dict[str, Any] = {}
        try:
            instrument_attributes = acq_data["instrument"]
        except KeyError:
            to_store["instrument"] = "NA"
            to_store["instrument_data"] = "NA"
        else:
            # Do some tidying of the instrument attributes
            instruments = set()
            cleaned_instrument_attributes = []

            if not isinstance(instrument_attributes, list):
                instrument_attributes = [instrument_attributes]

            for inst in instrument_attributes:
                instrument_name = inst["label"]
                instruments.add(instrument_name)
                uri = inst["uri"]

                cleaned_instrument_attributes.extend([instrument_name, uri])

            if len(instruments) == 1:
                instrument = instruments.pop()
            else:
                instrument = "multiple"

            to_store["instrument"] = instrument
            to_store["instrument_data"] = cleaned_instrument_attributes

        attributes.update(to_store)

        attributes["site"] = station_data["id"]
        attributes["measurement_type"] = measurement_type
        # TODO: Remove this from general attributes but make sure this is
        # included as a specific value on the appropriate variable.
        attributes["units"] = units

        _sampling_height = acq_data["samplingHeight"]
        attributes["sampling_height"] = format_inlet(_sampling_height, key_name="sampling_height")
        attributes["sampling_height_units"] = "metres"
        attributes["inlet"] = format_inlet(_sampling_height, key_name="inlet")
        attributes["inlet_height_magl"] = format_inlet(_sampling_height, key_name="inlet_height_magl")

        loc_data = station_data["location"]

        attributes["station_long_name"] = loc_data["label"]
        attributes["station_latitude"] = str(loc_data["lat"])
        attributes["station_longitude"] = str(loc_data["lon"])

        # 03/05/2023: Updated attributes to include altitude for "station_height_masl" explicitly.
        # attributes["station_altitude"] = format_inlet(loc_data["alt"], key_name="station_altitude")
        # attributes["station_height_masl"] = format_inlet(str(stat.eas), key_name="station_height_masl")
        attributes["station_height_masl"] = format_inlet(loc_data["alt"], key_name="station_height_masl")

        attributes["data_owner"] = f"{stat.firstName} {stat.lastName}"
        attributes["data_owner_email"] = str(stat.email)

        attributes["citation_string"] = dobj_info["references"]["citationString"]
        attributes["licence_name"] = dobj_info["references"]["licence"]["name"]
        attributes["licence_info"] = dobj_info["references"]["licence"]["url"]

        metadata = {}

        network = "ICOS"

        try:
            site_info = openghg_site_metadata[site.upper()][network]
        except KeyError:
            pass
        else:
            metadata["station_long_name"] = site_info["long_name"]
            metadata["station_latitude"] = site_info["latitude"]
            metadata["station_longitude"] = site_info["longitude"]

        # Add some values directly for attributes (for now)
        metadata["species"] = attributes["species"]

        # Add ICOS in directly here for now
        additional_data = {}
        additional_data["network"] = network
        additional_data["data_type"] = "surface"
        additional_data["data_source"] = "icoscp"
        additional_data["source_format"] = "icos"
        additional_data["icos_data_level"] = str(data_level)
        additional_data["dataset_source"] = dobj_dataset_source
        additional_data["site"] = site

        attributes.update(additional_data)
        metadata.update(additional_data)

        dataframe.columns = [x.lower() for x in dataframe.columns]
        dataframe = dataframe.dropna(axis="index")

        if not dataframe.index.is_monotonic_increasing:
            dataframe = dataframe.sort_index()

        spec = attributes["species"]

        rename_cols = {
            "stdev": spec + " variability",
            "nbpoints": spec + " number_of_observations",
        }

        # TODO - add this back in once we've merged the fixes in
        # Try and conver the flag / userflag column to str
        # possible_flag_cols = ("flag", "userflag")
        # flag_col = [x for x in dataframe.columns if x in possible_flag_cols]

        # PR328
        # if flag_col:
        #     flag_str = flag_col[0]
        #     dataframe = dataframe.astype({flag_str: str})

        dataframe = dataframe.rename(columns=rename_cols).set_index("timestamp")

        dataframe.index.name = "time"
        dataframe.index = to_datetime(dataframe.index, format="%Y-%m-%d %H:%M:%S")

        dataset = dataframe.to_xarray()
        dataset.attrs.update(attributes)

        # So there isn't an easy way of getting a hash of a Dataset, can we do something
        # simple here we can compare data that's being added? Then we'll be able to make sure
        # ObsSurface.store_data won't accept data it's already seen
        data_key = f"key-{n}"
        # TODO - do we need both attributes and metadata here?
        standardised_data[data_key] = {
            "metadata": metadata,
            "data": dataset,
            "attributes": attributes,
        }

    standardised_data = assign_attributes(data=standardised_data, update_mismatch=update_mismatch)

    return standardised_data


# def _read_site_metadata():
#     """ Read site metadata from object store, if it doesn't exist we'll
#     retrieve it from the ICOS CP and store it.

#     Returns:
#         dict: Dictionary of site data
#     """
#     from openghg.objectstore import get_bucket, get_object_from_json
#     from openghg.types import ObjectStoreError
#     from openghg.util import timestamp_now
# raise NotImplementedError
#     key = "metadata/icos_atmos_site_metadata"
#     bucket = get_bucket()

#     try:
#         data = get_object_from_json(bucket=bucket, key=key)
#     except ObjectStoreError:
#         # Retrieve and store
#         from icoscp import station
#         station_data = station.getIdList()
#         metadata = {d.id: dict(d) for _, d in df.iterrows()}



from pathlib import Path
from typing import Dict, Optional, Union
import warnings


def parse_openghg(
    filepath: Path,
    species: str,
    source: str,
    domain: str,
    data_type: str,
    database: Optional[str] = None,
    database_version: Optional[str] = None,
    model: Optional[str] = None,
    time_resolved: bool = False,
    high_time_resolution: bool = False,
    period: Optional[Union[str, tuple]] = None,
    chunks: Optional[Dict] = None,
    continuous: bool = True,
) -> Dict:
    """
    Read and parse input flux / emissions data already in OpenGHG format.

    Args:
        filepath: Path to the flux file.
        species: Name of species
        source: Source of the emissions data
        domain: Geographic domain
        data_type: Type of data
        database: Name of the database
        database_version: Version of the database
        model: Model name if applicable.
        time_resolved: If this is a high resolution file.
        high_time_resolution:  This argument is deprecated and will be replaced in future versions with time_resolved.
        period: The time period for which data is to be parsed.
        chunks: Chunking schema to use when storing data. It expects a dictionary of dimension name and chunk size,
            for example {"time": 100}. If None then a chunking schema will be set automatically by OpenGHG.
            See documentation for guidance on chunking: https://docs.openghg.org/tutorials/local/Adding_data/Adding_ancillary_data.html#chunking.
            To disable chunking pass in an empty dictionary.
        continuous: Flag indicating whether the data is continuous or not
    Returns:
        dict: Dictionary of data
    """
    from openghg.standardise.meta import assign_flux_attributes
    from openghg.store import infer_date_range, update_zero_dim
    from openghg.util import timestamp_now
    from xarray import open_dataset

    if high_time_resolution:
        warnings.warn(
            "This argument is deprecated and will be replaced in future versions with time_resolved.",
            DeprecationWarning,
        )
        time_resolved = high_time_resolution

    em_data = open_dataset(filepath).chunk(chunks)

    # Some attributes are numpy types we can't serialise to JSON so convert them
    # to their native types here
    attrs = {}
    for key, value in em_data.attrs.items():
        try:
            attrs[key] = value.item()
        except AttributeError:
            attrs[key] = value

    author_name = "OpenGHG Cloud"
    em_data.attrs["author"] = author_name

    metadata = {}
    metadata.update(attrs)

    metadata["species"] = species
    metadata["domain"] = domain
    metadata["source"] = source

    optional_keywords = {"database": database, "database_version": database_version, "model": model}
    for key, value in optional_keywords.items():
        if value is not None:
            metadata[key] = value

    metadata["author"] = author_name
    metadata["data_type"] = data_type
    metadata["processed"] = str(timestamp_now())
    metadata["data_type"] = "flux"
    metadata["source_format"] = "openghg"

    # As flux / emissions files handle things slightly differently we need to check the time values
    # more carefully.
    # e.g. a flux / emissions file could contain e.g. monthly data and be labelled as 2012 but
    # contain 12 time points labelled as 2012-01-01, 2012-02-01, etc.

    # Check if time has 0-dimensions and, if so, expand this so time is 1D
    if "time" in em_data.coords:
        em_data = update_zero_dim(em_data, dim="time")

    em_time = em_data["time"]

    start_date, end_date, period_str = infer_date_range(
        em_time, filepath=filepath, period=period, continuous=continuous
    )

    metadata["start_date"] = str(start_date)
    metadata["end_date"] = str(end_date)

    metadata["max_longitude"] = round(float(em_data["lon"].max()), 5)
    metadata["min_longitude"] = round(float(em_data["lon"].min()), 5)
    metadata["max_latitude"] = round(float(em_data["lat"].max()), 5)
    metadata["min_latitude"] = round(float(em_data["lat"].min()), 5)

    metadata["time_resolution"] = "high" if time_resolved else "standard"
    metadata["time_period"] = period_str

    key = "_".join((species, source, domain))

    flux_data: Dict[str, dict] = {}
    flux_data[key] = {}
    flux_data[key]["data"] = em_data
    flux_data[key]["metadata"] = metadata

    flux_data = assign_flux_attributes(flux_data)

    return flux_data



from abc import ABC, abstractmethod
from typing import Any, Optional
from xarray import Dataset
from collections.abc import Iterator


class Store(ABC):
    """Interface for storing data in a Datasource. This may be in a zarr directory
    store, compressed NetCDF, a sparse storage format or others."""

    @abstractmethod
    def add(
        self,
        version: str,
        dataset: Dataset,
        compressor: Optional[Any] = None,
        filters: Optional[Any] = None,
    ) -> None:
        """Add an xr.Dataset to the zarr store."""
        pass

    @abstractmethod
    def delete_version(self, version: str) -> None:
        """Delete a version from the store"""
        pass

    @abstractmethod
    def delete_all(self) -> None:
        """Remove data from the zarr store"""
        pass

    @abstractmethod
    def keys(self, version: str) -> Iterator[str]:
        """Keys of data stored in the zarr store"""
        pass

    @abstractmethod
    def close(self) -> None:
        """Close the zarr store."""
        pass

    @abstractmethod
    def store_key(self, version: str) -> str:
        """Return the key of this zarr store"""
        pass

    @abstractmethod
    def version_exists(self, version: str) -> bool:
        """Check if a version exists in the current store"""
        pass

    @abstractmethod
    def get(self, version: str) -> Dataset:
        """Get the version of the dataset stored in the zarr store."""
        pass

    @abstractmethod
    def _pop(self, version: str) -> Dataset:
        """Pop some data from the store. This removes the data at this version from the store
        and returns it."""
        pass

    @abstractmethod
    def update(
        self, version: str, dataset: Dataset, compressor: Optional[Any], filters: Optional[Any]
    ) -> None:
        """Update the data at the given key"""
        pass

    @abstractmethod
    def bytes_stored(self) -> int:
        """Return the number of bytes stored in the zarr store"""
        pass



from typing import Any, Dict, Optional, Iterable


def get_zarr_encoding(
    data_vars: Iterable, compressor: Optional[Any] = None, filters: Optional[Any] = None
) -> Dict:
    """Return a dictionary of zarr compression settings for the given data.

    Args:
        data_vars: Data variables to encode
        compressor: Compressor to use, see https://zarr.readthedocs.io/en/stable/tutorial.html#compressors
        filters: Filters to use, see https://zarr.readthedocs.io/en/stable/tutorial.html#filters
    Returns:
        Dict: Dictionary of encoding settings for zarr store
    """
    if compressor is None:
        return {}

    encoding = {"compressor": compressor}
    if filters is not None:
        encoding["filters"] = filters

    return {var: encoding for var in data_vars}



"""
Parse EDGAR flux files and perform area-conservative regridding
of fluxes.

Currently based on acrg.name.emissions_helperfuncs.getedgarannualtotals()

Additional edgar functions which could be incorporated.
- getedgarv5annualsectors
- getedgarv432annualsectors
- (monthly sectors?)

TODO: Work out how to select frequency
- could try and relate to period e.g. "monthly" versus "yearly" etc.

EDGAR flux files tend to have dimensions (lat, lon), and no time dimension;
however, time is included in the filename, so we try to extract this, along
with version, species, sector, and resolution.

Typical file names:

 v432_CH4_1978.0.1x0.1.nc (or .zip)
 v50_CH4_1978.0.1x0.1.nc (or .zip)
 v6.0_CH4_1978_TOTALS.0.1x0.1.nc

 v50_CO2_excl_short-cycle_org_C_1978.0.1x0.1.nc (or .zip)
 v50_CO2_org_short-cycle_C_1978.0.1x0.1.nc (or .zip)
 v50_N2O_1978.0.1x0.1.zip (or .zip)

Futher information is contained in the EDGAR readme files.
For instance, from "_readme.html" from v6.0 data:

'Yearly Emissions gridmaps in ton substance / 0.1degree x 0.1degree / year
 for the .txt files with longitude and latitude coordinates referring to
 the low-left corner of each grid-cell.'

'Monthly Emissions gridmaps in ton substance / 0.1degree x 0.1degree / month
 for the .txt files with longitude and latitude coordinates referring to
 the low-left corner of each grid-cell.'

'Emissions gridmaps in kg substance /m2 /s for the .nc files with longitude
 and latitude coordinates referring to the cell center of each grid-cell.'

For EDGAR v8.0,

"""

import pathlib
import re
import zipfile
from collections import namedtuple
from typing import Any, Dict, Optional, Tuple, Union, cast
import logging
import numpy as np
import xarray as xr
from numpy import ndarray

from openghg.standardise.meta import assign_flux_attributes, define_species_label
from openghg.store import infer_date_range
from openghg.util import (
    clean_string,
    molar_mass,
    synonyms,
    find_coord_name,
    convert_internal_longitude,
    timestamp_now,
)


logger = logging.getLogger("openghg.transform.flux")
logger.setLevel(logging.DEBUG)  # Have to set level for logger as well as handler


ArrayType = Optional[Union[ndarray, xr.DataArray]]


_edgar_known_versions = ("v432", "v50", "v60", "v70", "v80")


# TODO: make this work for...
# - yearly - "v6.0_CH4_2015_TOTALS.0.1x0.1.nc"
# - sectoral - "v6.0_CH4_2015_ENE.0.1x0.1.nc"
# - monthly sectoral - "v6.0_CH4_2015_1_ENE.0.1x0.1.nc", "v6.0_CH4_2015_2_ENE.0.1x0.1.nc", ...
def assemble_edgar_metadata(
    filepath: Union[pathlib.Path, zipfile.Path],
    species: Optional[str] = None,
    version: Optional[str] = None,
) -> dict[str, Any]:
    """Combine given metadata with metadata extracted from filename.

    Args:
        filepath: pathlib.Path or zipfile.Path to EDGAR datafile
        species: species of flux
        version: EDGAR database version

    Returns:
        dictionary containing keys: version, species, year, source, resolution
    """
    known_versions = _edgar_known_versions

    if species is not None:
        species = define_species_label(species)[0]

    valid_version = False
    if version is not None:
        version = clean_string(version)
        if version in known_versions:
            valid_version = True

    try:
        metadata = _extract_file_info(filepath)
    except ValueError as e:
        # NOTE: if species is not None and valid_version is True, then we have everything we
        # need except resolution at this point...
        raise ValueError(f"Could extract EDGAR metadata from file f{filepath.name}") from e
    else:
        if species:
            if species != synonyms(metadata["species"]):
                logger.warning(
                    "Input species does not match species extracted from database filenames. Please check."
                )
            metadata["species"] = species
        else:
            metadata["species"] = clean_string(metadata["species"])

        if valid_version:
            metadata["version"] = version
        else:
            if clean_string(metadata["version"]) not in known_versions:
                if version is not None:
                    raise ValueError(
                        f"Unable to infer EDGAR version ({version})."
                        f" Please pass as an argument (one of {known_versions})"
                    )
                else:
                    raise ValueError(
                        f"Unable to infer EDGAR version."
                        f" Please pass as an argument (one of {known_versions})"
                    )
            else:
                metadata["version"] = clean_string(metadata["version"])

        source_from_file = metadata["source"]
        if source_from_file in ("TOTALS", ""):
            source = "anthro"
        elif metadata["species"] == "co2" and "TOTALS" in source_from_file:
            co2_source = "_".join(source_from_file.split("_")[:-1])
            source = clean_string(f"{co2_source}_anthro")
        else:
            source = clean_string(source_from_file)

        metadata["source"] = source

    return metadata


def parse_edgar(
    datapath: pathlib.Path,
    date: str,
    species: Optional[str] = None,
    domain: Optional[str] = None,
    lat_out: ArrayType = None,
    lon_out: ArrayType = None,
    source: Optional[str] = None,
    edgar_version: Optional[str] = None,
) -> Dict:
    """
    Read and parse input EDGAR data.
    Notes: Only accepts annual 2D grid maps in netcdf (.nc) format for now.
           Does not accept monthly data yet.

    EDGAR data is global on a 0.1 x 0.1 grid. This function allows products
    to be created for a given year which cover specific regions (and matches
    to the OpenGHG data schema, including units and coordinate names).

    Region information can be specified as follows:
     - To use a pre-defined domain use the domain keyword only.
     - To define a new domain use the domain, lat_out, lon_out keywords
     - If no domain or lat_out, lon_out data is supplied, the global EDGAR
    data will be added labelled as "globaledgar" domain.

    Pre-exisiting domains are defined within the openghg_defs "domain_info.json" file.

    Metadata will also be added to the stored data including:
     - "domain": domain (e.g. "europe") OR "globaledgar"
     - "source": "anthro" (for "TOTAL"), source name from file otherwise
     - "database": "EDGAR"
     - "database_version": edgar_version (e.g. "v60", "v50", "v432")

    Args:
        datapath: Path to data folder or zip archive for EDGAR data
        date: Year to extract. Expect a string of the form "YYYY"
        species: Species name being extracted
        domain: Domain name for new or pre-existing domain
        lat_out: Latitude values for new domain
        lon_out: Longitude values for new domain
        source: Flux source to use; overrides the source extracted from the filename.
        edgar_version: EDGAR version in file. Will be inferred otherwise.

    Returns:
        dict: Dictionary of data

    TODO: Allow date range to be extracted rather than year?
    TODO: Add monthly parsing and sector stacking options
    """
    period = None

    # TODO: Add check for period? Only monthly or yearly (or equivalent inputs)

    if zipfile.is_zipfile(datapath):
        datapath = zipfile.Path(datapath)  # type: ignore

    # NOTE: the built-in `open` method doesn't work with zipfile.Path
    # but there is a zipfile.Path.open method that works the same as
    # pathlib.Path.open

    folder_filelist = [x for x in datapath.iterdir() if x.is_file()]

    # Extract netcdf files (only, for now) - ".txt" is also an option (not implemented)
    # Path(file.name).suffix workaround because zipfile.Path.suffix is Python 3.11+
    data_files = [file for file in folder_filelist if pathlib.Path(file.name).suffix == ".nc"]

    if not data_files:
        raise ValueError(f"No '.nc' files found in datapath: {datapath}")

    if edgar_version is None:
        # check if _readme.html is in folder_filelist, and if so, try to extract version
        readme = datapath / "_readme.html"
        if readme.exists():
            with readme.open("r") as f:  # pathlib.Path and zipfile.Path have .open method
                edgar_version = _check_readme_data(f.read())

    if len(date) == 4:
        year = int(date)
    else:
        raise ValueError(
            f"Date {date} does not represent a year;" " only annual EDGAR data can be processed currently."
        )

    FileInfo = namedtuple("FileInfo", "path metadata")
    files_by_year: dict[int, FileInfo] = {}
    for data_file in data_files:
        try:
            metadata = assemble_edgar_metadata(data_file, species, edgar_version)
        except ValueError:
            continue
        else:
            # Check if data is actually monthly "...2015_1" etc. - can't parse yet
            if "month" in metadata:
                raise NotImplementedError("Unable to parse monthly EDGAR data at present.")

            files_by_year[metadata["year"]] = FileInfo(data_file, metadata)

    if not files_by_year:
        raise ValueError(f"Unable to extract EDGAR file info from any files in {datapath}.")

    try:
        edgar_file, edgar_file_info = files_by_year[year]
    except KeyError:
        all_years = sorted(list(files_by_year.keys()))
        start_year, end_year = all_years[0], all_years[-1]

        if year < start_year:
            raise ValueError(f"Files span range: {start_year}-{end_year}." f" {year} is before this period.")
        elif year > end_year:
            logger.info(f"Using last available year from range:" f"{start_year}-{end_year}.")
        edgar_file, edgar_file_info = files_by_year[end_year]

    species_label = edgar_file_info["species"]
    version = edgar_file_info["version"]

    # get dataset
    # using .open("rb") for pathlib.Path and zipfile.Path compatibility
    edgar_ds = xr.open_dataset(edgar_file.open("rb"))

    # Expected name e.g. "emi_ch4", "emi_co2" for version <= 7; "fluxes" for version 8
    name = "fluxes" if version.startswith("v8") else f"emi_{species_label}"

    # check that data variable `name` is in `edgar_ds`
    if name not in edgar_ds.data_vars:
        if version.startswith("v8"):
            raise ValueError(
                f"Data variable {name} not present. We only support 'flx_nc' files, not 'emi_nc' files, for EDGAR v8.0"
            )
        else:
            raise ValueError(f"Data variable {name} not present.")

    # Convert from kg/m2/s to mol/m2/s
    species_molar_mass = molar_mass(species_label)
    kg_to_g = 1e3

    flux_da = edgar_ds[name]
    flux_da = flux_da * kg_to_g / species_molar_mass
    units = "mol/m2/s"

    # TODO: some options for f-gases (.emi files) have different units...
    # need to catch this

    lat_name = find_coord_name(flux_da, options=["lat", "latitude"])
    lon_name = find_coord_name(flux_da, options=["lon", "longitude"])
    if lat_name is None or lon_name is None:
        raise ValueError(
            f"Could not find '{lat_name}' or '{lon_name}' in EDGAR file.\n"
            " Please check this is a 2D grid map."
        )

    # Check range of longitude values and convert to -180 - +180
    flux_da = convert_internal_longitude(
        flux_da, lon_name=lon_name
    )  # TODO is this creating NaNs for East Asia domain?

    lat_out, lon_out = _check_lat_lon(domain, lat_out, lon_out)

    if lat_out is not None and lon_out is not None:
        # Will produce import error if xesmf has not been installed.
        from openghg.transform import regrid_uniform_cc
        from openghg.util import cut_data_extent

        # To improve performance of regridding algorithm cut down the data
        # to match the output grid (with buffer).
        flux_da_cut = cut_data_extent(flux_da, lat_out, lon_out)
        flux_values = flux_da_cut.values

        lat_in_cut = flux_da_cut[lat_name]
        lon_in_cut = flux_da_cut[lon_name]

        # area conservative regrid
        flux_values = regrid_uniform_cc(flux_values, lat_out, lon_out, lat_in_cut, lon_in_cut)
    else:
        lat_out = flux_da[lat_name]
        lon_out = flux_da[lon_name]
        flux_values = flux_da.values

    edgar_attrs = edgar_ds.attrs

    # Check for "time" dimension and add if missing.
    flux_ndim = flux_values.ndim
    time_name = "time"
    if time_name in flux_da:
        time = flux_da[time_name].values
        flux = flux_values  # TODO: this was missing... is this correct?? otherwise 'flux' might be undefined
    elif time_name not in flux_da and flux_ndim == 2:
        time = np.array([f"{year}-01-01"], dtype="datetime64[ns]")
        flux = flux_values[np.newaxis, ...]
    else:
        raise ValueError(
            f"Expected data variable '{name}' to contain 2 or 3 dimensions (including time),"
            f" but '{name}' has {flux_ndim} dimensions: {flux_da.dims}."
        )

    dims = ("time", "lat", "lon")

    em_data = xr.Dataset(
        {"flux": (dims, flux)}, coords={"time": time, "lat": lat_out, "lon": lon_out}, attrs=edgar_attrs
    )

    # Some attributes are numpy types we can't serialise to JSON so convert them
    # to their native types here
    attrs = {}
    for key, value in em_data.attrs.items():
        try:
            attrs[key] = value.item()
        except AttributeError:
            attrs[key] = value

    author_name = "OpenGHG Cloud"
    em_data.attrs["author"] = author_name

    metadata = {}
    metadata.update(attrs)

    raw_edgar_domain = "globaledgar"
    if domain is None:
        domain = raw_edgar_domain

    source = source if source is not None else edgar_file_info["source"]
    metadata["species"] = species_label
    metadata["domain"] = domain
    metadata["source"] = source
    metadata["database"] = "EDGAR"
    metadata["database_version"] = edgar_file_info["version"]
    metadata["author"] = author_name
    metadata["processed"] = str(timestamp_now())
    metadata["data_type"] = "flux"

    attrs = {"author": metadata["author"], "processed": metadata["processed"]}

    # Infer the date range associated with the flux data
    em_time = em_data.time
    start_date, end_date, period_str = infer_date_range(em_time, filepath=edgar_file.name, period=period)

    prior_info_dict = {
        "EDGAR": {
            "version": f"EDGAR {edgar_version}",
            "filename": edgar_file.name,
            "raw_resolution": "0.1 degrees x 0.1 degrees",
            "reference": edgar_ds.attrs["source"],
        }
    }

    metadata["start_date"] = str(start_date)
    metadata["end_date"] = str(end_date)

    metadata["min_longitude"] = round(float(em_data["lon"].min()), 5)
    metadata["max_longitude"] = round(float(em_data["lon"].max()), 5)
    metadata["min_latitude"] = round(float(em_data["lat"].min()), 5)
    metadata["max_latitude"] = round(float(em_data["lat"].max()), 5)

    metadata["time_resolution"] = "standard"
    metadata["time_period"] = period_str

    key = "_".join((species_label, source, domain, date))

    emissions_data: Dict[str, dict] = {}
    emissions_data[key] = {}
    emissions_data[key]["data"] = em_data
    emissions_data[key]["metadata"] = metadata
    emissions_data[key]["attributes"] = attrs

    emissions_data = assign_flux_attributes(emissions_data, units=units, prior_info_dict=prior_info_dict)

    return emissions_data


def _check_lat_lon(
    domain: Optional[str] = None, lat_out: ArrayType = None, lon_out: ArrayType = None
) -> Tuple[Optional[ndarray], Optional[ndarray]]:
    """
    Define and check latitude and longitude values for a domain.

    The domain can be used in one of two ways:
        1. To specify a pre-exisiting lat, lon extent which can be extracted
        2. To supply a name for a new lat, lon extent which must be specified

    For case 1, only domain needs to be specified (lat_out and lon_out can
    be specified but they must already exactly match the domain definition).
    The details will be extracted from openghg_defs "domain_info.json" file.

    For case 2, the domain, lat_out and lon_out must all be specified.

    If none of these values are specified, (None, None) will be returned. This
    is valid behaviour.

    Args:
        domain: Domain name for a pre-existing domain or a new domain
        lat_out: Latitude values (only needed if domain is new)
        lon_out: Longitude values (only needed if domain is new)

    Returns:
        ndarray, ndarray: Latitude and longitude arrays
        None, None: if all inputs are None, a tuple of Nones will be returned.
    """
    from openghg.util import convert_lon_to_360, find_domain

    if lat_out is not None or lon_out is not None:
        if domain is None:
            raise ValueError(
                "Please specify new 'domain' name if selecting new" " latitude, longitude values"
            )

    if isinstance(lat_out, xr.DataArray):
        lat_out = cast(ndarray, lat_out.values)

    if isinstance(lon_out, xr.DataArray):
        lon_out = cast(ndarray, lon_out.values)

    if domain is not None:
        # If domain is specified, attempt to extract lat/lon values from
        # pre-defined definitions.
        try:
            lat_domain, lon_domain = find_domain(domain)
        except ValueError:
            # If domain cannot be found and lat, lon values have not been
            # defined raise an error.
            if lat_out is None or lon_out is None:
                raise ValueError("To create new domain please input" " 'lat_out' and 'lon_out' values.")
        else:
            # Check domain latitude and longitude values against any
            # lat_out and lon_out values specified to check they match.
            if lat_out is not None:
                if not np.array_equal(lat_domain, lat_out):
                    raise ValueError(
                        "Latitude values should not be specified"
                        f" when using pre-defined domain {domain}"
                        " (values don't match)."
                    )
            else:
                lat_out = lat_domain

            if lon_out is not None:
                if not np.array_equal(lon_domain, lon_out):
                    raise ValueError(
                        "Longitude values should not be specified"
                        f" when using pre-defined domain {domain}"
                        " (values don't match)."
                    )
            else:
                lon_out = lon_domain

            if lon_out.max() - lon_out.min() >= 360:
                raise ValueError(
                    "Invalid domain definition."
                    " Expected longitude in range: 0 to 360."
                    f"Current longitude: {lon_out.min()} - {lon_out.max()}"
                )

    if lon_out is not None and (lon_out.max() > 180 or lon_out.min() < -180):
        logger.info("Converting longitude to stay within 0 to 360 bounds")
        lon_converted = convert_lon_to_360(lon_out)
        lon_out = cast(Optional[ndarray], lon_converted)

    return lat_out, lon_out


def _check_readme_data(readme_data: str) -> Optional[str]:
    """Parse EDGAR _readme.html to find version.

    Args:
        readme_data: string containing contents of _readme.html

    Returns:
        EDGAR version, if found, otherwise None
    """
    try:
        # Ignoring types as issues caught by try-except statement
        # Find and extract title line from html file
        title_line = re.search("<title.*?>(.+?)</title>", readme_data).group()  # type: ignore
        # Extract version e.g. "v6.0" or "v4.3.2"
        edgar_version = re.search(r"v\d[.]\d[.]?\d*", title_line).group()  # type: ignore
    except (ValueError, AttributeError):
        return None

        # Check against known versions and remove '.' if these don't match.
    if edgar_version not in _edgar_known_versions:
        edgar_version = edgar_version.replace(".", "")

    return edgar_version


def _extract_file_info(edgar_file: Union[pathlib.Path, zipfile.Path, str]) -> Dict:
    """
    Extract details from EDGAR filename.

    Expected filenames roughly of the form:
        - {version}_{species}_{year}.{resolution}.nc
        - {version}_{species}_{year}_{source}.{resolution}.nc
        - {version}_{species}_{year}_{month}.._{source..}.{resolution}.nc

    Args:
        edgar_file: Filename for an EDGAR database file

    Returns:
        Dict : Elements of the filename as a dictionary

        >>> _extract_file_info("v6.0_CH4_2015_TOTALS.0.1x0.1.nc")
            {"version": "v6.0", "species": "CH4", "year": 2015,
            "source": "TOTALS", "resolution": "0.1x0.1"}
        >>> _extract_file_info("v50_CH4_2015.0.1x0.1.nc")
            {"version": "v50", "species": "CH4", "year": 2015,
            "source": "", "resolution": "0.1x0.1"}
        >>> _extract_file_info("v432_CH4_2010_9_IPCC_6A_6D.0.1x0.1.nc")
            {"version": "v432", "species": "CH4", "year": 2010, "month": 9,
            "source": "IPCC-6A-6D", "resolution": "0.1x0.1"}
    """
    # Can extract details about files from the filename itself.
    # e.g. "v6.0_CH4_2015_TOTALS.0.1x0.1.nc"

    # Extract filename stem (name without extension) and split
    if isinstance(edgar_file, str):
        edgar_file = pathlib.Path(edgar_file)
    if isinstance(edgar_file, zipfile.Path):
        edgar_file = pathlib.Path(edgar_file.name)  # zipfile.Path.stem is Python 3.11+
    filename = edgar_file.stem

    # remove unwanted parts like:
    # FT2021, FT2022, etc
    # GHG
    # emi_nc, flx_nc
    cleaning_regexps = [r"FT(19|20)\d{2}_", r"GHG_", r"_(emi|flx)(_nc)?"]
    cleaning_regexps_joined = "|".join(cleaning_regexps)
    cleaning_pat = re.compile(rf"({cleaning_regexps_joined})")

    filename_clean = cleaning_pat.sub("", filename)

    # parse string of form:
    # {version}_{species}_{optional co2 info}_{year}_{optional month}_{optional sector}_{optional resolution}
    info_pat = re.compile(
        r"^(?P<version>v\d[\.\d]*)_"  # capture version e.g. v432, v50, v8.0
        r"(?P<species>[a-zA-Z\d-]+)_"  # capture species, e.g. CH4, c-C4F8, HFC-43-10-mee
        r"((?P<co2_options>(excl|org)_short-cycle)(_org)?(_C)?_)?"  # optional, capture CO2 info
        r"(?P<year>(19|20)\d{2})"  # capture year (might not end in _)
        r"(_(?P<month>\d{1,2}))?"  # optionally capture month
        r"(_(?P<sector>(\w+)))?"  # optionally capture sector, e.g. TOTALS, TNR_Ship, N2O, IPCC_4C_4D1_4D4
        r"(\.(?P<resolution>\d\.\dx\d\.\d))?"  # optionally capture resolution, e.g. 0.1x0.1
    )

    if m := info_pat.search(filename_clean):
        file_info = m.groupdict()
    else:
        # info_pat matches all files in known EDGAR versions
        # (verified by searching all file names for these versions
        # 18 March 2024)
        raise ValueError(f"Did not recognise input file format: {filename}")

    # make "source" string
    co2_options = file_info.pop("co2_options") or ""

    sector = file_info.pop("sector") or ""
    sector = sector.replace("_", "-")

    source = "_".join([co2_options, sector])
    source = source.strip("_")  # in case co2_options or sector is ""

    # if not source:
    #     raise ValueError(f"Unable to extract source: {filename}")

    file_info["source"] = source

    # year and month should be integers
    file_info["year"] = int(file_info["year"])

    if file_info["month"] is not None:
        file_info["month"] = int(file_info["month"])
    else:
        del file_info["month"]

    # file_info["version"] = clean_string(file_info["version"])

    # if not file_info["version"] in _edgar_known_versions:
    #     raise ValueError(f"Version {file_info['version']} inferred from {filename} not know known "
    #                      f"EDGAR versions: {_edgar_known_versions}.")

    return file_info


# def getedgarv5annualsectors(year, lon_out, lat_out, edgar_sectors, species='CH4'):
#     """
#     Get annual emission totals for species of interest from EDGAR v5.0 data
#     for sector or sectors.
#     Regrids to the desired lats and lons.

#     CURRENTLY ONLY 2012 AND 2015 ANNUAL SECTORS IN SHARED DIRECTORY. OTHER YEARS NEED DOWNLOADING.

#     Args:
#         year (int):
#             Year of interest
#         lon_out (array):
#             Longitudes to output the data on
#         lat_out (array):
#             Latitudes to output the data on
#         edgar_sectors (list of str) (optional):
#             EDGAR sectors to include. If list of values, the sum of these will be used.
#             See below for list of possible sectors and full names.
#         species (str):
#             Which species you want to look at.
#             e.g. species = 'CH4'
#             Default = 'CH4'
#             Currently only works for CH4.

#     Returns:
#         narr (array):
#             Array of regridded emissions in mol/m2/s.
#             Dimensions are [lat, lon]

#     If there is no data for the species you are looking at you may have to
#     download it from:
#     https://edgar.jrc.ec.europa.eu/overview.php?v=50_GHG
#     and place in:
#     /data/shared/Gridded_fluxes/<species>/EDGAR_v5.0/yearly_sectoral/

#     Note:
#         EDGAR sector names:
#         "AGS" = Agricultural soils
#         "AWB" = Agricultural waste burning
#         "CHE" = Chemical processes
#         "ENE" = Power industry
#         "ENF" = Enteric fermentation
#         "FFF" = Fossil fuel fires
#         "IND" = Combustion for manufacturing
#         "IRO" = Iron and steel production
#         "MNM" = Maure management
#         "PRO_COAL" = Fuel exploitation - coal
#         "PRO_GAS" = Fuel exploitation - gas
#         "PRO_OIL" = Fuel expoitation - oil
#         "PRO" = Fuel exploitation - contains coal, oil, gas
#         "RCO" = Energy for buildings
#         "REF_TRF" = Oil refineries and transformational industries
#         "SWD_INC" = Solid waste disposal - incineration
#         "SWD_LDF" = Solid waste disposal - landfill
#         "TNR_Aviation_CDS" = Aviation - climbing and descent
#         "TNR_Aviation_CRS" = Aviation - cruise
#         "TNR_Aviation_LTO" = Aviation - landing and takeoff
#         "TNR_Other" = Railways, pipelines and off-road transport
#         "TNR_Ship" = Shipping
#         "TRO" = Road transportation
#         "WWT" = Waste water treatment

#     """

#     edgarfp = os.path.join(data_path,"Gridded_fluxes",species.upper(),"EDGAR_v5.0/yearly_sectoral")

#     EDGARsectorlist = ["AGS","AWB","CHE","ENE","ENF","FFF","IND","IRO","MNM",
#                        "PRO_COAL","PRO_GAS","PRO_OIL","PRO","RCO","REF_TRF","SWD_INC",
#                        "SWD_LDF","TNR_Aviation_CDS","TNR_Aviation_CRS",
#                        "TNR_Aviation_LTO","TNR_Other","TNR_Ship","TRO","WWT"]

#     if edgar_sectors is not None:
#         print('Including EDGAR sectors.')

#         for EDGARsector in edgar_sectors:
#             if EDGARsector not in EDGARsectorlist:
#                 print('EDGAR sector {0} not one of: \n {1}'.format(EDGARsector,EDGARsectorlist))
#                 print('Returning None')
#                 return None

#         #edgar flux in kg/m2/s
#         for i,sector in enumerate(edgar_sectors):

#             edgarfn = "v50_" + species.upper() + "_" + str(year) + "_" + sector + ".0.1x0.1.nc"

#             with xr.open_dataset(os.path.join(edgarfp,edgarfn)) as edgar_file:
#                 edgar_flux = np.nan_to_num(edgar_file['emi_'+species.lower()].values,0.)
#                 edgar_lat = edgar_file.lat.values
#                 edgar_lon = edgar_file.lon.values

#             if i == 0:
#                 edgar_total = edgar_flux
#             else:
#                 edgar_total = np.add(edgar_total,edgar_flux)

#         edgar_regrid_kg,arr = regrid2d(edgar_total,edgar_lat,edgar_lon,lat_out,lon_out)

#         #edgar flux in mol/m2/s
#         speciesmm = molar_mass(species)
#         edgar_regrid = (edgar_regrid_kg.data*1e3) / speciesmm

#     return(edgar_regrid)

# def getedgarv432annualsectors(year, lon_out, lat_out, edgar_sectors, species='CH4'):
#     """
#     Get annual emission totals for species of interest from EDGAR v4.3.2 data
#     for sector or sectors.
#     Regrids to the desired lats and lons.

#     If there is no data for the species you are looking at you may have to
#     download it from:
#     http://edgar.jrc.ec.europa.eu/overview.php?v=432_GHG&SECURE=123
#     and placed in:
#     /data/shared/Gridded_fluxes/<species>/EDGAR_v4.3.2/<species>_sector_yearly/

#     Args:
#         year (int):
#             Year of interest
#         lon_out (array):
#             Longitudes to output the data on
#         lat_out (array):
#             Latitudes to output the data on
#         edgar_sectors (list):
#             List of strings of EDGAR sectors to get emissions for.
#             These will be combined to make one array.
#             See 'Notes' for names of sectors
#         species (str):
#             Which species you want to look at.
#             e.g. species = 'CH4'
#             Default = 'CH4'

#     Returns:
#         narr (array):
#             Array of regridded emissions in mol/m2/s.
#             Dimensions are [lat, lon]

#     Notes:
#         Names of EDGAR sectors:
#            'powerindustry';
#            'oilrefineriesandtransformationindustry';
#            'combustionformanufacturing';
#            'aviationclimbinganddescent';
#            'aviationcruise';
#            'aviationlandingandtakeoff';
#            'aviationsupersonic';
#            'roadtransport';
#            'railwayspipelinesandoffroadtransport';
#            'shipping';
#            'energyforbuildings';
#            'fuelexploitation';
#            'nonmetallicmineralsproduction';
#            'chemicalprocesses';
#            'ironandsteelproduction';
#            'nonferrousmetalsproduction';
#            'nonenergyuseoffuels';
#            'solventsandproductsuse';
#            'entericfermentation';
#            'manuremanagement';
#            'agriculturalsoils';
#            'indirectN2Oemissionsfromagriculture';
#            'agriculturalwasteburning';
#            'solidwastelandfills';
#            'wastewaterhandling';
#            'Solid waste incineration';
#            'fossilfuelfires';
#            'indirectemissionsfromNOxandNH3';
#     """
#     species = species.upper() #Make sure species is uppercase

# #Path to EDGAR files
#     edpath = os.path.join(data_path,'Gridded_fluxes/'+species+'/EDGAR_v4.3.2/'+species+'_sector_yearly/')

#     #Dictionary of codes for sectors
#     secdict = {'powerindustry' : '1A1a',
#                'oilrefineriesandtransformationindustry' : '1A1b_1A1c_1A5b1_1B1b_1B2a5_1B2a6_1B2b5_2C1b',
#                'combustionformanufacturing' : '1A2',
#                'aviationclimbinganddescent' : '1A3a_CDS',
#                'aviationcruise' : '1A3a_CRS',
#                'aviationlandingandtakeoff' : '1A3a_LTO',
#                'aviationsupersonic' : '1A3a_SPS',
#                'roadtransport' : '1A3b',
#                'railwayspipelinesandoffroadtransport' : '1A3c_1A3e',
#                'shipping' : '1A3d_1C2',
#                'energyforbuildings' : '1A4',
#                'fuelexploitation' : '1B1a_1B2a1_1B2a2_1B2a3_1B2a4_1B2c',
#                'nonmetallicmineralsproduction' : '2A',
#                'chemicalprocesses': '2B',
#                'ironandsteelproduction' : '2C1a_2C1c_2C1d_2C1e_2C1f_2C2',
#                'nonferrousmetalsproduction' : '2C3_2C4_2C5',
#                'nonenergyuseoffuels' : '2G',
#                'solventsandproductsuse' :  '3',
#                'entericfermentation' : '4A',
#                'manuremanagement' : '4B',
#                'agriculturalsoils' : '4C_4D',
#                'indirectN2Oemissionsfromagriculture' : '4D3',
#                'agriculturalwasteburning' : '4F',
#                'solidwastelandfills' : '6A_6D',
#                'wastewaterhandling' : '6B',
#                'Solid waste incineration' : '6C',
#                'fossilfuelfires' : '7A',
#                'indirectemissionsfromNOxandNH3' : '7B_7C'
#     }

#     #Check to see range of years. If desired year falls outside of this range
#     #then take closest year
#     possyears = np.empty(shape=[0,0],dtype=int)
#     for f in glob.glob(edpath+'v432_'+species+'_*'):
#         fname = f.split('/')[-1]
#         fyear = fname[9:13]      #Extract year from filename
#         possyears = np.append(possyears, int(fyear))
#     if year > max(possyears):
#         print("%s is later than max year in EDGAR database" % str(year))
#         print("Using %s as the closest year" % str(max((possyears))))
#         year = max(possyears)
#     if year < min(possyears):
#         print("%s is earlier than min year in EDGAR database" % str(year))
#         print("Using %s as the closest year" % str(min((possyears))))
#         year = min(possyears)


#     #Species molar mass
#     speciesmm = molar_mass(species)
# #    if species == 'CH4':
# #        #speciesmm = 16.0425
# #        speciesmm = molar_mass(species)
# #    elif species == 'N2O':
# #        speciesmm = 44.013
# #    else:
# #        print "No molar mass for species %s." % species
# #        print "Please add this and rerun the script"
# #        print "Returning None"
# #        return(None)


#     #Read in EDGAR data of annual mean CH4 emissions for each sector
#     #These are summed together
#     #units are in kg/m2/s
#     tot = None
#     for sec in edgar_sectors:
#         edgar = edpath+'v432_'+species+'_'+str(year)+'_IPCC_'+secdict[sec]+'.0.1x0.1.nc'
#         if os.path.isfile(edgar):
#             ds = xr.open_dataset(edgar)
#             soiname = 'emi_'+species.lower()
#             if tot is None:
#                 tot = ds[soiname].values*1e3 / speciesmm
#             else:
#                 tot += ds[soiname].values*1e3 / speciesmm
#         else:
#             print('No annual file for sector %s and %s' % (sec, species))

#     lat_in = ds.lat.values
#     lon_in = ds.lon.values

#     nlat = len(lat_out)
#     nlon = len(lon_out)

#     narr = np.zeros((nlat, nlon))
#     narr, reg = regrid2d(tot, lat_in, lon_in,
#                              lat_out, lon_out)

#     return(narr)

# def getedgarmonthlysectors(lon_out, lat_out, edgar_sectors, months=[1,2,3,4,5,6,7,8,9,10,11,12],
#                            species='CH4'):
#     """
#     Get 2010 monthly emissions for species of interest from EDGAR v4.3.2 data
#     for sector or sectors.
#     Regrids to the desired lats and lons.
#     If there is no data for the species you are looking at you may have to
#     download it from:
#     http://edgar.jrc.ec.europa.eu/overview.php?v=432_GHG&SECURE=123
#     and place it in:
#     /data/shared/Gridded_fluxes/<species>/EDGAR_v4.3.2/<species>_sector_monthly/

#     Args:
#         lon_out (array):
#             Longitudes to output the data on
#         lat_out (array):
#             Latitudes to output the data on
#         edgar_sectors (list):
#             List of strings of EDGAR sectors to get emissions for.
#             These will be combined to make one array.
#             See 'Notes' for names of sectors
#         months (list of int; optional):
#             Desired months.
#         species (str, optional):
#             Which species you want to look at.
#             e.g. species = 'CH4'
#             Default = 'CH4'

#     Returns:
#         narr (array):
#             Array of regridded emissions in mol/m2/s.
#             Dimensions are [no of months, lat, lon]

#     Notes:
#         Names of EDGAR sectors:
#            'powerindustry';
#            'oilrefineriesandtransformationindustry';
#            'combustionformanufacturing';
#            'aviationclimbinganddescent';
#            'aviationcruise';
#            'aviationlandingandtakeoff';
#            'aviationsupersonic';
#            'roadtransport';
#            'railwayspipelinesandoffroadtransport';
#            'shipping';
#            'energyforbuildings';
#            'fuelexploitation';
#            'nonmetallicmineralsproduction';
#            'chemicalprocesses';
#            'ironandsteelproduction';
#            'nonferrousmetalsproduction';
#            'nonenergyuseoffuels';
#            'solventsandproductsuse';
#            'entericfermentation';
#            'manuremanagement';
#            'agriculturalsoils';
#            'indirectN2Oemissionsfromagriculture';
#            'agriculturalwasteburning';
#            'solidwastelandfills';
#            'wastewaterhandling';
#            'Solid waste incineration';
#            'fossilfuelfires';
#            'indirectemissionsfromNOxandNH3';
#     """
#     species = species.upper() #Make sure species is uppercase
#     #Path to EDGAR files
#     edpath = os.path.join(data_path,'Gridded_fluxes/'+species+'/EDGAR_v4.3.2/'+species+'_sector_monthly/')

#     #Dictionary of codes for sectors
#     secdict = {'powerindustry' : '1A1a',
#                'oilrefineriesandtransformationindustry' : '1A1b_1A1c_1A5b1_1B1b_1B2a5_1B2a6_1B2b5_2C1b',
#                'combustionformanufacturing' : '1A2',
#                'aviationclimbinganddescent' : '1A3a_CDS',
#                'aviationcruise' : '1A3a_CRS',
#                'aviationlandingandtakeoff' : '1A3a_LTO',
#                'aviationsupersonic' : '1A3a_SPS',
#                'roadtransport' : '1A3b',
#                'railwayspipelinesandoffroadtransport' : '1A3c_1A3e',
#                'shipping' : '1A3d_1C2',
#                'energyforbuildings' : '1A4',
#                'fuelexploitation' : '1B1a_1B2a1_1B2a2_1B2a3_1B2a4_1B2c',
#                'nonmetallicmineralsproduction' : '2A',
#                'chemicalprocesses': '2B',
#                'ironandsteelproduction' : '2C1a_2C1c_2C1d_2C1e_2C1f_2C2',
#                'nonferrousmetalsproduction' : '2C3_2C4_2C5',
#                'nonenergyuseoffuels' : '2G',
#                'solventsandproductsuse' :  '3',
#                'entericfermentation' : '4A',
#                'manuremanagement' : '4B',
#                'agriculturalsoils' : '4C_4D',
#                'indirectN2Oemissionsfromagriculture' : '4D3',
#                'agriculturalwasteburning' : '4F',
#                'solidwastelandfills' : '6A_6D',
#                'wastewaterhandling' : '6B',
#                'Solid waste incineration' : '6C',
#                'fossilfuelfires' : '7A',
#                'indirectemissionsfromNOxandNH3' : '7B_7C'
#     }

#     print('Note that the only year for monthly emissions is 2010 so using that.')

#     #Species molar mass
#     speciesmm = molar_mass(species)
# #    if species == 'CH4':
# #        speciesmm = 16.0425
# #    elif species == 'N2O':
# #        speciesmm = 44.013
# #    else:
# #        print "No molar mass for species %s." % species
# #        print "Please add this and rerun the script"
# #        print "Returning None"
# #        return(None)


#     #Read in EDGAR data of annual mean CH4 emissions for each sector
#     #These are summed together
#     #units are in kg/m2/s
#     warnings = []
#     first = 0
#     for month in months:
#         tot = np.array(None)
#         for sec in edgar_sectors:
#             edgar = edpath+'v432_'+species+'_2010_'+str(month)+'_IPCC_'+secdict[sec]+'.0.1x0.1.nc'
#             if os.path.isfile(edgar):
#                 ds = xr.open_dataset(edgar)
#                 soiname = 'emi_'+species.lower()
#                 if tot.any() == None:
#                     tot = ds[soiname].values*1e3 / speciesmm
#                 else:
#                     tot += ds[soiname].values*1e3 / speciesmm
#             else:
#                 warnings.append('No monthly file for sector %s' % sec)
#                 #print 'No monthly file for sector %s' % sec

#             if first == 0:
#                 emissions = np.zeros((len(months), tot.shape[0], tot.shape[1]))
#                 emissions[0,:,:] = tot
#             else:
#                 first += 1
#                 emissions[first,:,:] = tot

#     for warning in np.unique(warnings):
#         print(warning)

#     lat_in = ds.lat.values
#     lon_in = ds.lon.values

#     nlat = len(lat_out)
#     nlon = len(lon_out)

#     narr = np.zeros((nlat, nlon, len(months)))

#     for i in range(len(months)):
#        narr[:,:,i], reg = regrid2d(emissions[i,:,:], lat_in, lon_in,
#                              lat_out, lon_out)
#     return(narr)
